/* ARCADEmini USB driver for Linux
 * version 1.2.2 date 2005-12-15
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2005 Stefan Schuermans <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/wrapper.h>
#include <linux/slab.h>
#include <linux/usb.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/atomic.h>

#include "am_usb.h"



//#define DEBUG



//version information
#define am_usb_name "ARCADEmini USB driver for Linux"
#define am_usb_ver "version 1.2.2 date 2005-12-15"

//#error new_ver: UPDATE numerical version information!!!
#define am_usb_ver_major 1
#define am_usb_ver_minor 2
#define am_usb_ver_revision 2
#define am_usb_ver_year 2005
#define am_usb_ver_month 12
#define am_usb_ver_day 15


//number of elements in an array
#define count( array ) (sizeof( (array) ) / sizeof( (array)[0] ))



//info macros
#define am_usb_info( text ) printk( KERN_INFO "am_usb: "text"\n" )
#define am_usb_infof( format, arg... ) printk( KERN_INFO "am_usb: "format"\n", ##arg )
//warning macros
#define am_usb_warning( text ) printk( KERN_WARNING "am_usb: "text"\n" )
#define am_usb_warningf( format, arg... ) printk( KERN_WARNING "am_usb: "format"\n", ##arg )
//error macros
#define am_usb_error( text ) printk( KERN_ERR "am_usb: "text"\n" )
#define am_usb_errorf( format, arg... ) printk( KERN_ERR "am_usb: "format"\n", ##arg )
//debug macros
#ifdef DEBUG
#define am_usb_debug( text ) printk( KERN_DEBUG "am_usb: "text"\n" )
#define am_usb_debugf( format, arg... ) printk( KERN_DEBUG "am_usb: "format"\n", ##arg )
#else
#define am_usb_debug( text )
#define am_usb_debugf( format, arg... )
#endif


//global constants concerning the device
#define am_usb_id_vendor 0x6666 //vendor id: prototype device
#define am_usb_id_product 0x416D //product id: 'Am' for ARCADEmini
//global constants concerning the module
#define am_usb_minor_base 0xA0 //base of minor device number
#define am_usb_minor_range 0x10 //range of minor device numbers
#define am_usb_devfs_node_name "am_usb" //node name for devfs



#ifdef DEBUG
//#define DEBUG_MEM
#endif
#ifdef DEBUG_MEM
//memory debug
struct semaphore my_alloc_lock;
int my_alloc_cnt = 0;
void * my_kmalloc( int len, int flags )
{
  void * ptr;
  down( &my_alloc_lock );
  ptr = kmalloc( len, flags );
  if( ptr )
  {
    my_alloc_cnt++;
    printk( KERN_INFO "ptr=%p alloc_cnt=%d\n", ptr, my_alloc_cnt );
  }
  up( &my_alloc_lock );
  return ptr;
}
void my_kfree( void * ptr )
{
  down( &my_alloc_lock );
  kfree( ptr );
  my_alloc_cnt--;
  printk( KERN_INFO "ptr=%p alloc_cnt=%d\n", ptr, my_alloc_cnt );
  up( &my_alloc_lock );
}
#define kmalloc my_kmalloc
#define kfree my_kfree
#endif



//module parameter - overall default mode
char * mode = "";
MODULE_PARM( mode, "s" );
int overall_default_mode = am_usb_mode_null;



//external variables
extern devfs_handle_t usb_devfs_handle; //handle of global usb directory of devfs



//a communication step
typedef struct t_am_usb_comm_step
{
  int out_len; //length of output buffer
  unsigned char * p_out_buffer; //buffer with data to send (GFP_KERNEL)
  int in_len; //length of input buffer
  unsigned char * p_filter_buffer; //buffer with filter for data to receive (GFP_KENREL, may be NULL)
  unsigned char * p_mask_buffer; //buffer with mask for data to receive (GFP_KERNEL, may be NULL)
  unsigned char * p_in_buffer; //buffer for received data (GFP_KERNEL, may be NULL)
} st_am_usb_comm_step;

//a communication block
typedef struct t_am_usb_comm_block
{
  struct t_am_usb_comm_block * p_next, * p_prev; //doubly linked list
  int linked; //flag if currently linked in doubly linked list
  int step_cnt; //number of communication steps
  struct t_am_usb_comm_step * steps; //array with communication steps (GFP_KERNEL)
  int out_step; //next step to output
  int in_step; //next step to input
  int timeout; //set to != 0 if timed out
  wait_queue_head_t * p_wait_queue; //waitqueue to wake up after completion of this block
                                    //(this pointer is only a copy and is not used to free the wait queue)
                                    //(may be NULL)
  int free_flag; //if != 0, this communication block is automatically freed after processing
} st_am_usb_comm_block;



//type for private data of a file
typedef struct t_am_usb_file_priv
{
  struct t_am_usb_dev * p_dev; //pointer to associated device
  struct t_am_usb_file_priv * p_next, * p_prev; //pointer to next and previous file that have opened same device

  wait_queue_head_t poll_wait_queue; //waitqueue for poll-call

  int mode; //the mode of this file

  unsigned long sec_no; //the current sector (for CF mode)
} st_am_usb_file_priv;

//type for private data of a device
typedef struct t_am_usb_dev
{
  struct semaphore file_lock; //used to syncronize file operations on this device
  struct semaphore lock; //used to lock this structure

  struct usb_device * p_usb_dev; //saved parameters of probe function
  unsigned int interface_no;
  const struct usb_device_id * p_usb_dev_id;

  unsigned char dev_no_sub_minor; //subminor device number (minor relative to base)
  unsigned char dev_no_minor; //minor device number

  unsigned char out_endpoint_addr; //address of the bulk out endpoint
  int out_max_packet_size; //maximum size of packets for bulk out endpoint
  unsigned char * p_out_buffer; //pointer to buffer for bulk out
  struct urb * p_out_urb; //bulk out URB
  int out_urb_linked; //boolean flag, if bulk out URB is submitted/linked

  unsigned char in_endpoint_addr; //address of the bulk in endpoint
  int in_max_packet_size; //maximum size of packets for bulk in endpoint
  unsigned char * p_in_buffer; //pointer to buffer for bulk in
  struct urb * p_in_urb; //bulk in URB
  int in_urb_linked; //boolean flag, if bulk in URB is submitted/linked

  char devfs_node_name[64]; //name of devfs node
  devfs_handle_t h_devfs_node; //handle of devfs node

  int open_cnt; //numer of times this file is opened at the moment
  struct t_am_usb_file_priv file_priv_list; //anchor of list with files that have opened this device

  struct t_am_usb_comm_block comm_block_list; //anchor of list of communication blocks

  int timeout_timer_active; //if != 0 timeout timer is currently running
  struct timer_list timeout_timer; //timer for timeout of USB bulk communication

  int mode; //the mode to use for new files
} st_am_usb_dev;



//table with pointers to private device data of devices
//(indexed by subminor)
static struct t_am_usb_dev * am_usb_dev_table[am_usb_minor_range];
//lock for this table
static struct semaphore am_usb_dev_table_lock;

//prototype of task function (needed by declaration of task queue)
static void am_usb_task( void * p_ingnored );
//taskqueue to do some things that cannot be done in interrupt
static struct tq_struct am_usb_task_queue =
{
  list: { NULL },
  sync: 0, 
  routine: am_usb_task,
  data: NULL
};
//boolean flags for tasks to do in next am_usb_task, modified by atomic bitops
//(am_usb_task is task for things that cannot be done in interrupt)
static long am_usb_task_flags[am_usb_minor_range];
#define am_usb_task_flag_comm_block_send 0 //try to send next step of communication block
#define am_usb_task_flag_comm_block_recv 1 //post-process received steps of communication block



//*** support functions

//free a communication block
static void am_usb_comm_block_free( struct t_am_usb_comm_block * p_comm_block )
{
  int step;

  //communication block is still linked
  if( p_comm_block->linked )
  {
    //remove pointer to wait queue
    p_comm_block->p_wait_queue = NULL;
    //set free flag to free the block automatically after completion
    p_comm_block->free_flag = 1;
    //do nothing else
    return;
  }

  //free buffers of communication steps
  for( step = 0; step < p_comm_block->step_cnt; step++ )
  {
    if( p_comm_block->steps[step].p_out_buffer != NULL )
      kfree( p_comm_block->steps[step].p_out_buffer );
    if( p_comm_block->steps[step].p_filter_buffer != NULL )
      kfree( p_comm_block->steps[step].p_filter_buffer );
    if( p_comm_block->steps[step].p_mask_buffer != NULL )
      kfree( p_comm_block->steps[step].p_mask_buffer );
    if( p_comm_block->steps[step].p_in_buffer != NULL )
      kfree( p_comm_block->steps[step].p_in_buffer );
  }

  //free array with communication steps
  kfree( p_comm_block->steps );

  //free communication block
  kfree( p_comm_block );

  //debug message
  am_usb_debugf( "freed communication block %p", p_comm_block );
}



//*** bulk output interrupt functions

//callback-procedure for USB bulk output
//(interrupt context)
static void am_usb_out_callback( struct urb * p_urb )
{
  struct t_am_usb_dev * p_dev;
#ifdef DEBUG
  int i, j;
  char dbg_text[256];
#endif

  //get device
  p_dev = (struct t_am_usb_dev *)p_urb->context;

  //bulk out URB is no more submitted/linked
  p_dev->out_urb_linked = 0;

  //error
  if( p_urb->status != USB_ST_NOERROR )
    //print error message
    am_usb_errorf( "bulk output reported error: %d", p_urb->status );
  //no error
  else
  {
    //not all data was sent
    if( p_urb->actual_length != p_urb->transfer_buffer_length )
      //print warning message
      am_usb_warningf( "bulk output returned a different length: %d (expected: %d)",
          p_urb->actual_length, p_urb->transfer_buffer_length );
#ifdef DEBUG
    //print bulk message with data
    dbg_text[0] = 0;
    for( i = 0, j = 0; i < p_urb->actual_length && j < sizeof( dbg_text ) - 4; i++, j += 3 )
      sprintf( dbg_text + j, "%02X ", ((unsigned char *)p_urb->transfer_buffer)[i] );
    am_usb_debugf( "bulk output done (len=%d): %s", p_urb->actual_length, dbg_text );
#endif
  }

  //set flag for execution of am_usb_comm_block_send for this device
  //(use test_and_x here to be smp safe)
  test_and_set_bit( am_usb_task_flag_comm_block_send, &am_usb_task_flags[p_dev->dev_no_sub_minor] );
  //schedule task for things that can not be done in interrupt
  schedule_task( &am_usb_task_queue );
}



//*** bulk input interrupt functions

//process a bulk in message
//(may be called in interrupt context with device locked by someone else)
static void am_usb_in_msg( struct t_am_usb_dev * p_dev, unsigned char * p_msg, int msg_len )
{
  struct t_am_usb_comm_block * p_comm_block;
  int step, i;

  //no first communication block
  if( p_dev->comm_block_list.p_next == &p_dev->comm_block_list )
    return; //do nothing

  //get pointer to first communication block
  p_comm_block = p_dev->comm_block_list.p_next;

  //at least one input communication step left to do
  if( p_comm_block->in_step < p_comm_block->step_cnt )
  {
    //filter and mask buffer present
    step = p_comm_block->in_step;
    if( p_comm_block->steps[step].p_filter_buffer != NULL
     && p_comm_block->steps[step].p_mask_buffer != NULL )
    {
      //match message against next step
      if( p_comm_block->steps[step].in_len != msg_len ) //wrong length
        return; //do nothing
      for( i = 0; i < msg_len; i++ )
        if( p_comm_block->steps[step].p_filter_buffer[i] !=
            (p_msg[i] & p_comm_block->steps[step].p_mask_buffer[i]) )
          break; //byte does not match
      if( i < msg_len ) //some byte does not match
        return; //do nothing
    }

    //copy recevied message into input buffer
    if( p_comm_block->steps[step].p_in_buffer != NULL )
      memcpy( p_comm_block->steps[step].p_in_buffer, p_msg, msg_len );

    //increment number of current input step
    p_comm_block->in_step++;
  }

  //communication block done
  if( (p_comm_block->out_step >= p_comm_block->step_cnt
    && p_comm_block->in_step >= p_comm_block->step_cnt) )
  {
    //delete timeout timer
    if( p_dev->timeout_timer_active )
    {
      del_timer_sync( &p_dev->timeout_timer );
      p_dev->timeout_timer_active = 0;
    }

    //set flag for execution of am_usb_comm_block_recv for this device
    //(use test_and_x here to be smp safe)
    test_and_set_bit( am_usb_task_flag_comm_block_recv, &am_usb_task_flags[p_dev->dev_no_sub_minor] );
    //schedule task for things that can not be done in interrupt
    schedule_task( &am_usb_task_queue );
  }
}

//callback-procedure for USB bulk input
//(interrupt context)
static void am_usb_in_callback( struct urb * p_urb )
{
  struct t_am_usb_dev * p_dev;
#ifdef DEBUG
  int i, j;
  char dbg_text[256];
#endif

  //get device
  p_dev = (struct t_am_usb_dev *)p_urb->context;

  //bulk in URB is no more submitted/linked
  p_dev->in_urb_linked = 0;

  //error
  if( p_urb->status != USB_ST_NOERROR )
    //print error message
    am_usb_errorf( "bulk input reported error: %d", p_urb->status );
  //no error
  else
  {
#ifdef DEBUG
    //print bulk message with data
    dbg_text[0] = 0;
    for( i = 0, j = 0; i < p_urb->actual_length && j < sizeof( dbg_text ) - 4; i++, j += 3 )
      sprintf( dbg_text + j, "%02X ", ((unsigned char *)p_urb->transfer_buffer)[i] );
    am_usb_debugf( "bulk input done (len=%d): %s", p_urb->actual_length, dbg_text );
#endif

    //process the message
    //(device may be locked by someone else, we are in interrupt context)
    am_usb_in_msg( p_dev, (unsigned char *)p_urb->transfer_buffer, p_urb->actual_length );

    //set up data in bulk in URB
    FILL_BULK_URB( p_dev->p_in_urb, p_dev->p_usb_dev, usb_rcvbulkpipe( p_dev->p_usb_dev, p_dev->in_endpoint_addr ),
        p_dev->p_in_buffer, p_dev->in_max_packet_size, am_usb_in_callback,
        p_dev );

    //bulk in URB will be linked/submitted soon
    //(do this here because callback could perhaps be called from usb_submit_urb)
    p_dev->in_urb_linked = 1;
    //resubmit bulk in URB
    if( usb_submit_urb( p_dev->p_in_urb ) )
    {
      //bulk in URB is not linked/submitted
      p_dev->in_urb_linked = 0;
      //print error message
      am_usb_error( "could not submit bulk in URB" );
    }
    else
    {
      //print info that bulk in was started
      am_usb_debugf( "bulk input started (max_len=%d)", p_dev->p_in_urb->transfer_buffer_length );
    }
  }
}

//callback-procedure for USB bulk timeout
//(interrupt context)
static void am_usb_timeout( unsigned long ul_p_dev )
{
  struct t_am_usb_dev * p_dev = (struct t_am_usb_dev *)ul_p_dev;
  struct t_am_usb_comm_block * p_comm_block;

  //timeout timer no more active
  p_dev->timeout_timer_active = 0;
  
  //no first communication block
  if( p_dev->comm_block_list.p_next == &p_dev->comm_block_list )
    return; //do nothing

  //get pointer to first communication block
  p_comm_block = p_dev->comm_block_list.p_next;

  //debug message
  am_usb_debugf( "comunication block %p timed out", p_comm_block );

  //set timeout flag
  p_comm_block->timeout = 1;

  //set flag for execution of am_usb_comm_block_recv for this device
  //(use test_and_x here to be smp safe)
  test_and_set_bit( am_usb_task_flag_comm_block_recv, &am_usb_task_flags[p_dev->dev_no_sub_minor] );
  //schedule task for things that can not be done in interrupt
  schedule_task( &am_usb_task_queue );
}



//*** functions to do some things that cannot be done in interrupt

//try to send next step of communication block
//(must be called with device locked)
static void am_usb_comm_block_send( struct t_am_usb_dev * p_dev )
{
  struct t_am_usb_comm_block * p_comm_block;
  int step, len;
#ifdef DEBUG
  int i, j;
  char dbg_text[256];
#endif

  //bulk out URB is already submitted/linked at the moment
  if( p_dev->out_urb_linked )
    return; //do nothing

  //no first communication block
  if( p_dev->comm_block_list.p_next == &p_dev->comm_block_list )
    return; //do nothing

  //get pointer to first communication block
  p_comm_block = p_dev->comm_block_list.p_next;

  //all output communication steps done
  if( p_comm_block->out_step >= p_comm_block->step_cnt )
    return; //do nothing

  //copy next step into bulk output buffer
  step = p_comm_block->out_step;
  len = min( p_comm_block->steps[step].out_len, p_dev->out_max_packet_size );
  memcpy( p_dev->p_out_buffer, p_comm_block->steps[step].p_out_buffer, len );

  //delete old timeout timer
  if( p_dev->timeout_timer_active )
  {
    del_timer_sync( &p_dev->timeout_timer );
    p_dev->timeout_timer_active = 0;
  }
  //create timeout timer
  init_timer( &p_dev->timeout_timer );
  p_dev->timeout_timer.expires = jiffies + max( HZ/50, 1 ) + 1;
  p_dev->timeout_timer.data = (unsigned long)p_dev;
  p_dev->timeout_timer.function = am_usb_timeout;
  add_timer( &p_dev->timeout_timer );
  p_dev->timeout_timer_active = 1;

  //increment number of current output step
  p_comm_block->out_step++;

  //set up data in bulk out URB
  FILL_BULK_URB( p_dev->p_out_urb, p_dev->p_usb_dev, usb_sndbulkpipe( p_dev->p_usb_dev, p_dev->out_endpoint_addr ),
    p_dev->p_out_buffer, len, am_usb_out_callback, p_dev );

  //bulk out URB will be linked/submitted soon
  //(do this here because callback could perhaps be called from usb_submit_urb)
  p_dev->out_urb_linked = 1;
  //submit bulk out URB
  if( usb_submit_urb( p_dev->p_out_urb ) )
  {
    //bulk out URB is not linked/submitted
    p_dev->out_urb_linked = 0;
    //print error message
    am_usb_error( "could not submit bulk out URB" );
  }
  else
  {
#ifdef DEBUG
    //print bulk message with data
    dbg_text[0] = 0;
    for( i = 0, j = 0; i < p_dev->p_out_urb->transfer_buffer_length && j < sizeof( dbg_text ) - 4; i++, j += 3 )
      sprintf( dbg_text + j, "%02X ", ((unsigned char *)p_dev->p_out_urb->transfer_buffer)[i] );
    am_usb_debugf( "bulk output started (len=%d): %s", p_dev->p_out_urb->transfer_buffer_length, dbg_text );
#endif
  }
}

//post-process received steps of communication block
//(must be called with device locked)
static void am_usb_comm_block_recv( struct t_am_usb_dev * p_dev )
{
  struct t_am_usb_comm_block * p_comm_block;

  //no first communication block
  if( p_dev->comm_block_list.p_next == &p_dev->comm_block_list )
    return; //do nothing

  //get pointer to first communication block
  p_comm_block = p_dev->comm_block_list.p_next;

  //communication block done
  if( (p_comm_block->out_step >= p_comm_block->step_cnt
    && p_comm_block->in_step >= p_comm_block->step_cnt)
   || p_comm_block->timeout )
  {
    //remove first communication block from list
    p_dev->comm_block_list.p_next = p_dev->comm_block_list.p_next->p_next;
    p_dev->comm_block_list.p_next->p_prev = &p_dev->comm_block_list;
    p_comm_block->linked = 0;

    //debug message
    am_usb_debugf( "completed communication block %p", p_comm_block );

    //wake up wait queue of this communication block
    if( p_comm_block->p_wait_queue != NULL )
      wake_up_interruptible( p_comm_block->p_wait_queue );

    //free communication block
    if( p_comm_block->free_flag )
      am_usb_comm_block_free( p_comm_block );

    //set flag for execution of am_usb_comm_block_send for this device
    //(use test_and_x here to be smp safe)
    test_and_set_bit( am_usb_task_flag_comm_block_send, &am_usb_task_flags[p_dev->dev_no_sub_minor] );
    //schedule task for things that can not be done in interrupt
    schedule_task( &am_usb_task_queue );
  }
}

//task to do some things that cannot be done in interrupt
static void am_usb_task( void * p_ingnored )
{
  int i;
  struct t_am_usb_dev * p_dev;

  //before waiting for any lock, peek at flags to see if we have got something to do
  for( i = 0; i < am_usb_minor_range; i++ )
  {
    if( test_bit( am_usb_task_flag_comm_block_send, &am_usb_task_flags[i] ) )
      break;
    if( test_bit( am_usb_task_flag_comm_block_recv, &am_usb_task_flags[i] ) )
      break;
  }
  //nothing to do
  if( i >= am_usb_minor_range )
    return;

  //lock device table
  down( &am_usb_dev_table_lock );

  //check all devices in device table
  for( i = 0; i < am_usb_minor_range; i++ )
  {
    //get pointer to device
    p_dev = am_usb_dev_table[i];
    //device is there
    if( p_dev != NULL )
    {
      //lock device
      down( &p_dev->lock );

      //device was not unplugged
      if( p_dev->p_usb_dev != NULL )
      {
        //call functions whose flag is set and clear the flag
        if( test_and_clear_bit( am_usb_task_flag_comm_block_send, &am_usb_task_flags[i] ) )
          am_usb_comm_block_send( p_dev );
        if( test_and_clear_bit( am_usb_task_flag_comm_block_recv, &am_usb_task_flags[i] ) )
          am_usb_comm_block_recv( p_dev );
      }

      //unlock the device
      up( &p_dev->lock );
    }
  }

  //unlock device table
  up( &am_usb_dev_table_lock );
}



//*** support functions for non-interrupt-time

//allocate a communication block
static struct t_am_usb_comm_block * am_usb_comm_block_new( int step_cnt, char * p_out_lens, char * p_in_lens )
{
  struct t_am_usb_comm_block * p_comm_block;
  int i;
  
  //invalid number of steps
  if( step_cnt <= 0 )
    return NULL;

  //allocate communication block
  p_comm_block = (struct t_am_usb_comm_block *)kmalloc( sizeof( struct t_am_usb_comm_block ), GFP_KERNEL );
  if( p_comm_block == NULL )
    return NULL;
  //initialize communication block
  p_comm_block->linked = 0;
  p_comm_block->step_cnt = step_cnt;
  p_comm_block->out_step = 0;
  p_comm_block->in_step = 0;
  p_comm_block->timeout = 0;
  p_comm_block->p_wait_queue = NULL;
  p_comm_block->free_flag = 0;

  //allocate communication steps
  p_comm_block->steps = (struct t_am_usb_comm_step *)kmalloc( step_cnt * sizeof( struct t_am_usb_comm_step ), GFP_KERNEL );
  if( p_comm_block->steps == NULL )
  {
    kfree( p_comm_block );
    return NULL;
  }
  //initialize steps
  for( i = 0; i < step_cnt; i++ )
  {
    p_comm_block->steps[i].out_len = max( 1, (int)(p_out_lens[i]) );
    p_comm_block->steps[i].p_out_buffer = NULL;
    p_comm_block->steps[i].in_len = max( 0, (int)(p_in_lens[i]) );
    p_comm_block->steps[i].p_filter_buffer = NULL;
    p_comm_block->steps[i].p_mask_buffer = NULL;
    p_comm_block->steps[i].p_in_buffer = NULL;
  }  
  
  //allocate buffers
  for( i = 0; i < step_cnt; i++ )
  {
    p_comm_block->steps[i].p_out_buffer = (unsigned char *)kmalloc( p_comm_block->steps[i].out_len, GFP_KERNEL );
    if( p_comm_block->steps[i].p_out_buffer == NULL )
    {
      am_usb_comm_block_free( p_comm_block );
      return NULL;
    }
    if( p_comm_block->steps[i].in_len > 0 )
    {
      p_comm_block->steps[i].p_filter_buffer = (unsigned char *)kmalloc( p_comm_block->steps[i].in_len, GFP_KERNEL );
      p_comm_block->steps[i].p_mask_buffer = (unsigned char *)kmalloc( p_comm_block->steps[i].in_len, GFP_KERNEL );
      p_comm_block->steps[i].p_in_buffer = (unsigned char *)kmalloc( p_comm_block->steps[i].in_len, GFP_KERNEL );
      if( p_comm_block->steps[i].p_filter_buffer == NULL
       || p_comm_block->steps[i].p_mask_buffer == NULL
       || p_comm_block->steps[i].p_in_buffer == NULL )
      {
        am_usb_comm_block_free( p_comm_block );
        return NULL;
      }
    }
  } //for( i ...

  //debug message
  am_usb_debugf( "new communication block %p", p_comm_block );

  //return pointer to communication block
  return p_comm_block;
}



//*** support functions for non-interrupt-time - device locked

//put communication block into communication block list
//(p_comm_block must be allocated with kmalloc( ..., GFP_KERNEL ))
//(must be called with device locked)
static void am_usb_comm_block_add( struct t_am_usb_dev * p_dev, struct t_am_usb_comm_block * p_comm_block )
{
  //do nothing if already linked in list
  if( p_comm_block->linked )
    return;

  //initialize communication block
  p_comm_block->out_step = 0;
  p_comm_block->in_step = 0;
  p_comm_block->timeout = 0;
  //put entry into list
  p_comm_block->p_prev = p_dev->comm_block_list.p_prev;
  p_comm_block->p_next = &p_dev->comm_block_list;
  p_comm_block->p_next->p_prev = p_comm_block;
  p_comm_block->p_prev->p_next = p_comm_block;
  p_comm_block->linked = 1;
  
  //debug message
  am_usb_debugf( "added communication block %p", p_comm_block );

  //set flag for execution of am_usb_comm_block_send for this device
  //(use test_and_x here to be smp safe)
  test_and_set_bit( am_usb_task_flag_comm_block_send, &am_usb_task_flags[p_dev->dev_no_sub_minor] );
  //schedule task for things that can not be done in interrupt
  schedule_task( &am_usb_task_queue );
}



//*** support functions for non-interrupt-time - device _not_ locked

//free a device and everything that belongs to it
//device is also removed from the device table
//(the device table must be locked)
static void am_usb_dev_free( struct t_am_usb_dev * p_dev )
{
  struct t_am_usb_comm_block * p_comm_block;

  //remove device from the device table
  am_usb_dev_table[p_dev->dev_no_sub_minor] = NULL;

  //delete timeout timer
  if( p_dev->timeout_timer_active )
    del_timer_sync( &p_dev->timeout_timer );

  //free URBs
  if( p_dev->p_out_urb != NULL )
  {
    if( p_dev->out_urb_linked )
      usb_unlink_urb( p_dev->p_out_urb );
    usb_free_urb( p_dev->p_out_urb );
  }
  if( p_dev->p_in_urb != NULL )
  {
    if( p_dev->in_urb_linked )
      usb_unlink_urb( p_dev->p_in_urb );
    usb_free_urb( p_dev->p_in_urb );
  }
  //free buffers
  if( p_dev->p_out_buffer != NULL )
    kfree( p_dev->p_out_buffer );
  if( p_dev->p_in_buffer != NULL )
    kfree( p_dev->p_in_buffer );

  //free communication blocks
  while( p_dev->comm_block_list.p_next != &p_dev->comm_block_list )
  {
    //remove first communication block from list
    p_comm_block = p_dev->comm_block_list.p_next;
    p_dev->comm_block_list.p_next = p_dev->comm_block_list.p_next->p_next;
    p_dev->comm_block_list.p_next->p_prev = &p_dev->comm_block_list;
    //free communication block
    if( p_comm_block->free_flag )
      am_usb_comm_block_free( p_comm_block );
  }

  //free device structure
  kfree( p_dev );
}

//process a communication block
//returns 0 on success
//(may not be called with device locked)
static int am_usb_comm_block_do( struct t_am_usb_dev * p_dev, struct t_am_usb_comm_block * p_comm_block )
{
  wait_queue_head_t wait_queue;
  int timeout_jiffies;

  //initialize wait queue to wait on
  init_waitqueue_head( &wait_queue );

  //initialize communication block wait queue and free_flag
  p_comm_block->p_wait_queue = &wait_queue;
  p_comm_block->free_flag = 0; //do not automatically free this communication block

  //lock device
  down( &p_dev->lock );
  
  //device was unplugged
  if( p_dev->p_usb_dev == NULL )
  {
    //unlock device
    up( &p_dev->lock );
    //no such device there
    return -ENODEV;
  }

  //add communication block to list
  am_usb_comm_block_add( p_dev, p_comm_block );

  //unlock device
  up( &p_dev->lock );
  
  //wait for communication block to be processed
  //sleep on wait queue
  timeout_jiffies = jiffies + p_comm_block->step_cnt * 2 * (max( HZ / 50, 1 ) + 1);
  interruptible_sleep_on_timeout( &wait_queue, timeout_jiffies );

  //check for errors
  if( p_comm_block->out_step < p_comm_block->step_cnt
   || p_comm_block->in_step < p_comm_block->step_cnt ) //communication block not completed
  {
    //timeout
    if( p_comm_block->timeout || jiffies >= timeout_jiffies )
      return -ETIMEDOUT;
    //error but not a timeout - interruptible_sleep_on_timeout interrupted by a signal
    return -EINTR;
  }

  //success
  return 0;
}



//*** file operations

//an userspace process is seeking on one of the devices
static loff_t am_usb_file_llseek( struct file * p_file, loff_t offset, int origin )
{
  struct t_am_usb_file_priv * p_file_priv;
  struct t_am_usb_dev * p_dev;
  struct t_am_usb_comm_block * p_comm_block;
  loff_t ret_val;
  unsigned long sec_cnt, new_sec_no;

  //get private file
  p_file_priv = (struct t_am_usb_file_priv *)p_file->private_data;
  //get private device
  p_dev = p_file_priv->p_dev;
  //no pointer to device
  if( p_dev == NULL )
    //so there is no device
    return -ENODEV;

  //lock device for file operations
  down( &p_dev->file_lock );

  //switch mode
  switch( p_file_priv->mode )
  {
    //act as compact flash reader / writer
    case am_usb_mode_cf:
      //get size
      //allocate communication block
      p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x05" );
      if( p_comm_block == NULL )
      {
        ret_val = -ENOMEM;
        break;
      }
      //fill communication block
      memcpy( p_comm_block->steps[0].p_out_buffer, "\x40", 1 );
      memcpy( p_comm_block->steps[0].p_filter_buffer, "\x40\0\0\0\0", 5 );
      memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0", 5 );
      //process commmunication block
      ret_val = am_usb_comm_block_do( p_dev, p_comm_block );
      if( ret_val != 0 )
      {
        am_usb_comm_block_free( p_comm_block );
        break;
      }
      sec_cnt = __be32_to_cpu( *(__u32 *)(&p_comm_block->steps[0].p_in_buffer[1]) );
      //free communication block
      am_usb_comm_block_free( p_comm_block );
      //adjust offset
      switch (origin)
      {
        //SEEK_END
        case 2:
          new_sec_no = sec_cnt + (unsigned long)(offset >> 9);
          break;
        //SEEK_CUR
        case 1:
          new_sec_no = p_file_priv->sec_no + (unsigned long)(offset >> 9);
          break;
        //SEEK_SET
        case 0:
        default:
          new_sec_no = (unsigned long)(offset >> 9);
      }
      //out of range
      if( new_sec_no > sec_cnt )
      {
        ret_val = -EINVAL;
        break;
      }
      //set new offset
      p_file_priv->sec_no = new_sec_no;
      ret_val = (loff_t)new_sec_no << 9;
      break; //case am_usb_mode_cf

    //act as null device
    case am_usb_mode_null:
    //show stream written to device
    case am_usb_mode_stream:
    default:
      //seek not possible
      ret_val = -EINVAL;
  } //switch( p_file_priv->mode )

  //debug message
  if( ret_val >= 0 )
    am_usb_debugf( "seek on file %08X (%d, %d) to position %lld (offset=%lld, origin=%d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val, offset, origin );
  else
    am_usb_debugf( "seek on file %08X (%d, %d) failed", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return new offset
  return ret_val;
}

//an userspace process is reading from one of the devices
static ssize_t am_usb_file_read( struct file * p_file, char * p_buffer, size_t count, loff_t * p_pos )
{
  struct t_am_usb_file_priv * p_file_priv;
  struct t_am_usb_dev * p_dev;
  struct t_am_usb_comm_block * p_comm_block;
  int i, cnt;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_buffer, count ) )
    //pointer not okay
    return -EFAULT;

  //get private file
  p_file_priv = (struct t_am_usb_file_priv *)p_file->private_data;
  //get private device
  p_dev = p_file_priv->p_dev;
  //no pointer to device
  if( p_dev == NULL )
    //so there is no device
    return -ENODEV;

  //lock device for file operations
  down( &p_dev->file_lock );

  //switch mode
  switch( p_file_priv->mode )
  {
    //act as compact flash reader / writer
    case am_usb_mode_cf:
      //buffer smaller than sector (512 bytes)
      if( count < 512 )
      {
        cnt = 0;
        break;
      }
      //read sector
      //allocate communication block
      p_comm_block = am_usb_comm_block_new( 17, "\x05\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02",
                                                "\x06\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22" );
      if( p_comm_block == NULL )
      {
        cnt = -ENOMEM;
        break;
      }
      //fill communication block
      p_comm_block->steps[0].p_out_buffer[0] = 0x43;
      *(__u32 *)(&p_comm_block->steps[0].p_out_buffer[1]) = __cpu_to_be32( p_file_priv->sec_no );
      memcpy( p_comm_block->steps[0].p_filter_buffer, "\x43\0\0\0\0\0", 6 );
      memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0\0", 6 );
      for( i = 0; i < 16; i++ )
      {
        p_comm_block->steps[i+1].p_out_buffer[0] = 0x31;
        p_comm_block->steps[i+1].p_out_buffer[1] = i;
        p_comm_block->steps[i+1].p_filter_buffer[0] = 0x31;
        p_comm_block->steps[i+1].p_filter_buffer[1] = i;
        memset( &p_comm_block->steps[i+1].p_filter_buffer[2], 0, 32 );
        p_comm_block->steps[i+1].p_mask_buffer[0] = 0xFF;
        p_comm_block->steps[i+1].p_mask_buffer[1] = 0xFF;
        memset( &p_comm_block->steps[i+1].p_mask_buffer[2], 0, 32 );
      }
      //process commmunication block
      cnt = am_usb_comm_block_do( p_dev, p_comm_block );
      if( cnt != 0 )
      {
        am_usb_comm_block_free( p_comm_block );
        break;
      }
      if( __be32_to_cpu( *(__u32 *)(&p_comm_block->steps[0].p_in_buffer[1]) ) != p_file_priv->sec_no
       || p_comm_block->steps[0].p_in_buffer[5] != 0 )
      {
        am_usb_comm_block_free( p_comm_block );
        cnt = -EIO;
        break;
      }
      //copy data to user space
      for( i = 0; i < 16; i++ )
        copy_to_user( &p_buffer[i << 5], &p_comm_block->steps[i+1].p_in_buffer[2], 32 );
      //free communication block
      am_usb_comm_block_free( p_comm_block );
      //update file pointer
      p_file_priv->sec_no++;
      //512 bytes read
      cnt = 512;
      break; //case am_usb_mode_cf

    //act as null device
    case am_usb_mode_null:
    //show stream written to device
    case am_usb_mode_stream:
    default:
      //nothing to read
      cnt = 0;
  } //switch( p_file_priv->mode )

  //debug message
  am_usb_debugf( "read from file %08X (%d, %d): %d bytes", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, cnt );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return number of bytes read
  return cnt;
}

//an userspace process is writing to one of the devices
static ssize_t am_usb_file_write( struct file * p_file, const char * p_buffer, size_t count, loff_t * p_pos )
{
  struct t_am_usb_file_priv * p_file_priv;
  struct t_am_usb_dev * p_dev;
  int i, cnt, pack, ofs, shift, channels, maxval, x, y, xx, yy;
  int fill_l, fill_r, fill_t, fill_b, skip_l, skip_t, startpos;
  unsigned long magic;
  unsigned short width, height;
  unsigned char val;
  struct t_am_usb_comm_block * p_comm_block;

  //test pointer
  if( verify_area( VERIFY_READ, p_buffer, count ) )
    //pointer not okay
    return -EFAULT;

  //get private file
  p_file_priv = (struct t_am_usb_file_priv *)p_file->private_data;
  //get private device
  p_dev = p_file_priv->p_dev;
  //no pointer to device
  if( p_dev == NULL )
    //so there is no device
    return -ENODEV;

  //lock device for file operations
  down( &p_dev->file_lock );

  //switch mode
  switch( p_file_priv->mode )
  {
    //show stream written to device
    case am_usb_mode_stream:
      //default: only one byte written
      // - idea is to skip only the first byte if the data could not be understood
      if( count >= 1 )
        cnt = 1;
      else
        cnt = 0;

      //get magic
      magic = 0;
      if( count >= 4 )
      {
        //get magic
        get_user( magic, (unsigned long *)p_buffer );
        magic = ntohl( magic );
      }
      //get parameters
      switch( magic )
      {
        //DEADBEEF and FEEDBEEF frame
        case 0xDEADBEEF:
        case 0xFEEDBEEF:
          //check length of packet
          if( count < 12 )
          {
            magic = 0;
            width = 0;
            height = 0;
            channels = 1;
            maxval = 1;
            startpos = 4;
            break;
          }
          //get width and height
          get_user( width, (unsigned short *)(p_buffer+8) );
          width = ntohs( width );
          get_user( height, (unsigned short *)(p_buffer+10) );
          height = ntohs( height );
          //set rest of parameters
          channels = 1;
          maxval = magic == 0xFEEDBEEF ? 0xFF : 1;
          startpos = 12;
          break;
        //MCU frame
        case 0x23542666:
          //check length of packet
          if( count < 12 )
          {
            magic = 0;
            width = 0;
            height = 0;
            channels = 1;
            maxval = 1;
            startpos = 4;
            break;
          }
          //get height and widtht
          get_user( height, (unsigned short *)(p_buffer+4) );
          height = ntohs( height );
          get_user( width, (unsigned short *)(p_buffer+6) );
          width = ntohs( width );
          //get channels and maxval
          get_user( channels, (unsigned short *)(p_buffer+8) );
          channels = ntohs( channels );
          get_user( maxval, (unsigned short *)(p_buffer+10) );
          maxval = ntohs( maxval );
          //set rest of parameters
          startpos = 12;
          break;
        default:
          magic = 0;
          width = 0;
          height = 0;
          channels = 1;
          maxval = 1;
          startpos = 4;
          break;
      }
      //some valid packet
      while( magic != 0 ) //use while command instead of if command to have break command
      {
        //check length of packet
        if( count < startpos + (int)width * (int)height * (int)channels )
          { magic = 0; break; }
        //calculate parameters
        fill_l = width < 26 ? (26 - width) / 2 : 0;
        fill_r = width < 26 ? 26 - width - fill_l : 0;
        fill_t = height < 20 ? (20 - height) / 2 : 0;
        fill_b = height < 20 ? 20 - height - fill_t : 0;
        skip_l = width > 26 ? (width - 26) / 2 : 0;
        skip_t = height > 20 ? (height - 20) / 2 : 0;
        //allocate communication block
        p_comm_block = am_usb_comm_block_new( 6, "\x36\x36\x36\x36\x36\x01", "\0\0\0\0\0\0" );
        if( p_comm_block == NULL )
          break;
        //fill communication block
        for( pack = 0; pack < 5; pack++ )
        {
          p_comm_block->steps[pack].p_out_buffer[0] = 0x10;
          p_comm_block->steps[pack].p_out_buffer[1] = (unsigned char)pack;
          for( i = 2; i < 54; i++ )
            p_comm_block->steps[pack].p_out_buffer[i] = 0x00;
        }
        for( y = 0; y < 20; y++ )
        for( x = 0; x < 26; x++ )
        {
          //get packet and offset
          i = y * 13 + x / 2;
          pack = i / 52;
          ofs = i % 52 + 2;
          shift = (1 - x % 2) * 4;
          //real pixel
          if( x >= fill_l && x < 26 - fill_r && y >= fill_t && y < 20 - fill_b )
          {
            xx = x - fill_l + skip_l;
            yy = y - fill_t + skip_t;
            get_user( val, p_buffer + startpos + (yy * width + xx) * channels );
            val = val > maxval ? 0x0F : (unsigned char)((unsigned long)(0x0F * val) / maxval);
            p_comm_block->steps[pack].p_out_buffer[ofs] |= val << shift;
          }
        }
        //fill show packet with data
        p_comm_block->steps[5].p_out_buffer[0] = 0x20;
        //lock device
        down( &p_dev->lock );
        //device was unplugged
        if( p_dev->p_usb_dev == NULL )
        {
          //free communication block
          am_usb_comm_block_free( p_comm_block );
          //unlock device
          up( &p_dev->lock );
          //unlock device for file operations
          up( &p_dev->file_lock );
          //no such device there
          return -ENODEV;
        }
        //submit the communication block
        p_comm_block->p_wait_queue = NULL; //do not signal completion
        p_comm_block->free_flag = 1; //automatically free after processing
        am_usb_comm_block_add( p_dev, p_comm_block );
        //unlock device
        up( &p_dev->lock );
        //return number of bytes written to this device
        cnt = startpos + (int)width * (int)height * (int)channels;
        //we used the while command only to have the break command
        //we wanted an if command
        break;
      }
      break; //case am_usb_mode_stream

    //act as compact flash reader / writer
    case am_usb_mode_cf:
      //no data to write
      if( count == 0 )
      {
        cnt = 0;
        break;
      }
      //write sector
      //allocate communication block
      p_comm_block = am_usb_comm_block_new( 17, "\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x05",
                                                "\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x06" );
      if( p_comm_block == NULL )
      {
        cnt = -ENOMEM;
        break;
      }
      //fill communication block
      for( i = 0; i < 16; i++ )
      {
        p_comm_block->steps[i].p_out_buffer[0] = 0x30;
        p_comm_block->steps[i].p_out_buffer[1] = i;
        memset( &p_comm_block->steps[i].p_out_buffer[2], 0, 32 );
        if( count >= (i << 5) + 32 )
          copy_from_user( &p_comm_block->steps[i].p_out_buffer[2], &p_buffer[i << 5], 32 );
        else if( count > i << 5 )
          copy_from_user( &p_comm_block->steps[i].p_out_buffer[2], &p_buffer[i << 5], count - (i << 5) );
        p_comm_block->steps[i].p_filter_buffer[0] = 0x30;
        p_comm_block->steps[i].p_filter_buffer[1] = i;
        p_comm_block->steps[i].p_mask_buffer[0] = 0xFF;
        p_comm_block->steps[i].p_mask_buffer[1] = 0xFF;
      }
      p_comm_block->steps[16].p_out_buffer[0] = 0x42;
      *(__u32 *)(&p_comm_block->steps[16].p_out_buffer[1]) = __cpu_to_be32( p_file_priv->sec_no );
      memcpy( p_comm_block->steps[16].p_filter_buffer, "\x42\0\0\0\0\0", 6 );
      memcpy( p_comm_block->steps[16].p_mask_buffer, "\xFF\0\0\0\0\0", 6 );
      //process commmunication block
      cnt = am_usb_comm_block_do( p_dev, p_comm_block );
      if( cnt != 0 )
      {
        am_usb_comm_block_free( p_comm_block );
        break;
      }
      if( __be32_to_cpu( *(__u32 *)(&p_comm_block->steps[16].p_in_buffer[1]) ) != p_file_priv->sec_no
       || p_comm_block->steps[16].p_in_buffer[5] != 0 )
      {
        am_usb_comm_block_free( p_comm_block );
        cnt = -EIO;
        break;
      }
      //free communication block
      am_usb_comm_block_free( p_comm_block );
      //update file pointer
      p_file_priv->sec_no++;
      //512 bytes written (if 512 bytes were supplied)
      cnt = min( (int)count, 512 );
      break; //case am_usb_mode_cf

    //act as null device
    case am_usb_mode_null:
    default:
      //write everything
      cnt = count;
  } //switch( p_file_priv->mode )

  //debug message
  am_usb_debugf( "wrote to file %08X (%d, %d): %d bytes", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, cnt );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return number of bytes written to this device
  return cnt;
}

//an userspace process is doing a poll or a select on one of the devices
static unsigned int am_usb_file_poll( struct file * p_file, poll_table * p_poll_table )
{
  struct t_am_usb_file_priv * p_file_priv;
  struct t_am_usb_dev * p_dev;
  unsigned int flags;

  //get private file
  p_file_priv = (struct t_am_usb_file_priv *)p_file->private_data;
  //get private device
  p_dev = p_file_priv->p_dev;
  //no pointer to device
  if( p_dev == NULL )
    //no device there
    return POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM | POLLHUP | POLLERR;

  //lock device for file operations
  down( &p_dev->file_lock );
  //lock device
  down( &p_dev->lock );

  //device was unplugged
  if( p_dev->p_usb_dev == NULL )
  {
    //unlock device
    up( &p_dev->lock );
    //unlock device for file operations
    up( &p_dev->file_lock );
    //no device there
    return POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM | POLLHUP | POLLERR;
  }

  //add our waitqueue to the poll table
  poll_wait( p_file, &p_file_priv->poll_wait_queue, p_poll_table );

  //switch mode
  switch( p_file_priv->mode )
  {
    //act as compact flash reader / writer
    case am_usb_mode_cf:
      //readable, writeable, no error
      flags = POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM;
      break;

    //show stream written to device
    case am_usb_mode_stream:
      //readable, writeable, no error
      flags = POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM;
      break;
    
    //act as null device
    case am_usb_mode_null:
    default:
      //readable, writeable, no error
      flags = POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM;
  } //switch( p_file_priv->mode )

  //debug message
  am_usb_debugf( "poll on file %08X (%d, %d): in=%d, out=%d, err=%d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, (flags & POLLIN) != 0, (flags & POLLOUT) != 0, (flags & POLLERR) != 0 );

  //unlock device
  up( &p_dev->lock );
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return flags
  return flags;
}

//halt ARCADEmini
//turns off command processing
static int am_usb_file_ioctl_halt( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x01" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x00;
  p_comm_block->steps[0].p_filter_buffer[0] = 0x00;
  p_comm_block->steps[0].p_mask_buffer[0] = 0xFF;
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl halt on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//run ARCADEmini
//turns on command processing
static int am_usb_file_ioctl_run( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x01" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x01;
  p_comm_block->steps[0].p_filter_buffer[0] = 0x01;
  p_comm_block->steps[0].p_mask_buffer[0] = 0xFF;
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl run on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//step ARCADEmini
//processes a single command
static int am_usb_file_ioctl_step( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x01" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x02;
  p_comm_block->steps[0].p_filter_buffer[0] = 0x02;
  p_comm_block->steps[0].p_mask_buffer[0] = 0xFF;
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl step on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//set command pointer
static int am_usb_file_ioctl_set_cmd_ptr( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_set_cmd_ptr * p_set_cmd_ptr )
{
  struct t_am_usb_set_cmd_ptr set_cmd_ptr;
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //test pointer
  if( verify_area( VERIFY_READ, p_set_cmd_ptr, sizeof( struct t_am_usb_set_cmd_ptr ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &set_cmd_ptr, p_set_cmd_ptr, sizeof( struct t_am_usb_set_cmd_ptr ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x07", "\x07" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x08;
  *(__u32 *)(&p_comm_block->steps[0].p_out_buffer[1]) = __cpu_to_be32( set_cmd_ptr.sec_no );
  *(__u16 *)(&p_comm_block->steps[0].p_out_buffer[5]) = __cpu_to_be16( set_cmd_ptr.ofs );
  p_comm_block->steps[0].p_filter_buffer[0] = 0x08;
  memset( &p_comm_block->steps[0].p_filter_buffer[1], 0, 6 );
  p_comm_block->steps[0].p_mask_buffer[0] = 0xFF;
  memset( &p_comm_block->steps[0].p_mask_buffer[1], 0, 6 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    set_cmd_ptr.sec_no = __be32_to_cpu( *(__u32 *)(&p_comm_block->steps[0].p_in_buffer[1]) );
    set_cmd_ptr.ofs = __be16_to_cpu( *(__u16 *)(&p_comm_block->steps[0].p_in_buffer[5]) );
    //copy structure to user-space
    copy_to_user( p_set_cmd_ptr, &set_cmd_ptr, sizeof( struct t_am_usb_set_cmd_ptr ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl set_cmd_ptr on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get command pointer
static int am_usb_file_ioctl_get_cmd_ptr( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_cmd_ptr * p_get_cmd_ptr )
{
  struct t_am_usb_get_cmd_ptr get_cmd_ptr;
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_cmd_ptr, sizeof( struct t_am_usb_get_cmd_ptr ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x07" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x09;
  p_comm_block->steps[0].p_filter_buffer[0] = 0x09;
  memset( &p_comm_block->steps[0].p_filter_buffer[1], 0, 6 );
  p_comm_block->steps[0].p_mask_buffer[0] = 0xFF;
  memset( &p_comm_block->steps[0].p_mask_buffer[1], 0, 6 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    get_cmd_ptr.sec_no = __be32_to_cpu( *(__u32 *)(&p_comm_block->steps[0].p_in_buffer[1]) );
    get_cmd_ptr.ofs = __be16_to_cpu( *(__u16 *)(&p_comm_block->steps[0].p_in_buffer[5]) );
    //copy structure to user-space
    copy_to_user( p_get_cmd_ptr, &get_cmd_ptr, sizeof( struct t_am_usb_get_cmd_ptr ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_cmd_ptr on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//load pixel buffer
//loads the pixel buffer with 260 bytes of data
static int am_usb_file_ioctl_load_pix_buf( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_load_pix_buf * p_load_pix_buf )
{
  struct t_am_usb_comm_block * p_comm_block;
  int i, ret_val;

  //test pointer
  if( verify_area( VERIFY_READ, p_load_pix_buf, sizeof( struct t_am_usb_load_pix_buf ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 5, "\x36\x36\x36\x36\x36", "\x02\x02\x02\x02\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  for( i = 0; i < 5; i++ )
  {
    p_comm_block->steps[i].p_out_buffer[0] = 0x10;
    p_comm_block->steps[i].p_out_buffer[1] = i;
    copy_from_user( &p_comm_block->steps[i].p_out_buffer[2], &p_load_pix_buf->pixels[i * 52], 52 );
    p_comm_block->steps[i].p_filter_buffer[0] = 0x10;
    p_comm_block->steps[i].p_filter_buffer[1] = i;
    p_comm_block->steps[i].p_mask_buffer[0] = 0xFF;
    p_comm_block->steps[i].p_mask_buffer[1] = 0xFF;
  }
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl load_pix_buf on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//read pixel buffer
//reads 260 bytes of data from the pixel buffer
static int am_usb_file_ioctl_read_pix_buf( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_read_pix_buf * p_read_pix_buf )
{
  struct t_am_usb_comm_block * p_comm_block;
  int i, ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_read_pix_buf, sizeof( struct t_am_usb_read_pix_buf ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 5, "\x02\x02\x02\x02\x02", "\x36\x36\x36\x36\x36" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  for( i = 0; i < 5; i++ )
  {
    p_comm_block->steps[i].p_out_buffer[0] = 0x11;
    p_comm_block->steps[i].p_out_buffer[1] = i;
    p_comm_block->steps[i].p_filter_buffer[0] = 0x11;
    p_comm_block->steps[i].p_filter_buffer[1] = i;
    memset( &p_comm_block->steps[i].p_filter_buffer[2], 0, 52 );
    p_comm_block->steps[i].p_mask_buffer[0] = 0xFF;
    p_comm_block->steps[i].p_mask_buffer[1] = 0xFF;
    memset( &p_comm_block->steps[i].p_mask_buffer[2], 0, 52 );
  }
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //copy data to user-space structure
    for( i = 0; i < 5; i++ )
      copy_to_user( &p_read_pix_buf->pixels[i * 52], &p_comm_block->steps[i].p_in_buffer[2], 52 );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl read_pix_buf on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//load overlay pixel buffer
//loads the overlay pixel buffer with 260 bytes of data
static int am_usb_file_ioctl_load_ovp_buf( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_load_ovp_buf * p_load_ovp_buf )
{
  struct t_am_usb_comm_block * p_comm_block;
  int i, ret_val;

  //test pointer
  if( verify_area( VERIFY_READ, p_load_ovp_buf, sizeof( struct t_am_usb_load_ovp_buf ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 5, "\x36\x36\x36\x36\x36", "\x02\x02\x02\x02\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  for( i = 0; i < 5; i++ )
  {
    p_comm_block->steps[i].p_out_buffer[0] = 0x12;
    p_comm_block->steps[i].p_out_buffer[1] = i;
    copy_from_user( &p_comm_block->steps[i].p_out_buffer[2], &p_load_ovp_buf->pixels[i * 52], 52 );
    p_comm_block->steps[i].p_filter_buffer[0] = 0x12;
    p_comm_block->steps[i].p_filter_buffer[1] = i;
    p_comm_block->steps[i].p_mask_buffer[0] = 0xFF;
    p_comm_block->steps[i].p_mask_buffer[1] = 0xFF;
  }
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl load_ovp_buf on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//read overlay pixel buffer
//reads 260 bytes of data from the overlay pixel buffer
static int am_usb_file_ioctl_read_ovp_buf( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_read_ovp_buf * p_read_ovp_buf )
{
  struct t_am_usb_comm_block * p_comm_block;
  int i, ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_read_ovp_buf, sizeof( struct t_am_usb_read_ovp_buf ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 5, "\x02\x02\x02\x02\x02", "\x36\x36\x36\x36\x36" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  for( i = 0; i < 5; i++ )
  {
    p_comm_block->steps[i].p_out_buffer[0] = 0x13;
    p_comm_block->steps[i].p_out_buffer[1] = i;
    p_comm_block->steps[i].p_filter_buffer[0] = 0x13;
    p_comm_block->steps[i].p_filter_buffer[1] = i;
    memset( &p_comm_block->steps[i].p_filter_buffer[2], 0, 52 );
    p_comm_block->steps[i].p_mask_buffer[0] = 0xFF;
    p_comm_block->steps[i].p_mask_buffer[1] = 0xFF;
    memset( &p_comm_block->steps[i].p_mask_buffer[2], 0, 52 );
  }
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //copy data to user-space structure
    for( i = 0; i < 5; i++ )
      copy_to_user( &p_read_ovp_buf->pixels[i * 52], &p_comm_block->steps[i].p_in_buffer[2], 52 );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl read_ovp_buf on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//swap display buffers
//exchanges the foreground and background pixel buffer
static int am_usb_file_ioctl_swap_bufs( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x01" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x20;
  p_comm_block->steps[0].p_filter_buffer[0] = 0x20;
  p_comm_block->steps[0].p_mask_buffer[0] = 0xFF;
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl swap_bufs on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//swap display buffers and copy image
//exchanges the foreground and background pixel buffer
//then copies the foreground picture into the background buffer
static int am_usb_file_ioctl_swap_bufs_cp( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x01" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x21;
  p_comm_block->steps[0].p_filter_buffer[0] = 0x21;
  p_comm_block->steps[0].p_mask_buffer[0] = 0xFF;
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl swap_bufs_cp on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//load sector buffer
//loads the sector buffer with 512 bytes of data
static int am_usb_file_ioctl_load_sec_buf( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_load_sec_buf * p_load_sec_buf )
{
  struct t_am_usb_comm_block * p_comm_block;
  int i, ret_val;

  //test pointer
  if( verify_area( VERIFY_READ, p_load_sec_buf, sizeof( struct t_am_usb_load_sec_buf ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 16, "\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22",
                                            "\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  for( i = 0; i < 16; i++ )
  {
    p_comm_block->steps[i].p_out_buffer[0] = 0x30;
    p_comm_block->steps[i].p_out_buffer[1] = i;
    copy_from_user( &p_comm_block->steps[i].p_out_buffer[2], &p_load_sec_buf->data[i << 5], 32 );
    p_comm_block->steps[i].p_filter_buffer[0] = 0x30;
    p_comm_block->steps[i].p_filter_buffer[1] = i;
    p_comm_block->steps[i].p_mask_buffer[0] = 0xFF;
    p_comm_block->steps[i].p_mask_buffer[1] = 0xFF;
  }
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl load_sec_buf on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//read sector buffer
//reads 512 bytes of data from the sector buffer
static int am_usb_file_ioctl_read_sec_buf( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_read_sec_buf * p_read_sec_buf )
{
  struct t_am_usb_comm_block * p_comm_block;
  int i, ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_read_sec_buf, sizeof( struct t_am_usb_read_sec_buf ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 16, "\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02",
                                            "\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  for( i = 0; i < 16; i++ )
  {
    p_comm_block->steps[i].p_out_buffer[0] = 0x31;
    p_comm_block->steps[i].p_out_buffer[1] = i;
    p_comm_block->steps[i].p_filter_buffer[0] = 0x31;
    p_comm_block->steps[i].p_filter_buffer[1] = i;
    memset( &p_comm_block->steps[i].p_filter_buffer[2], 0, 32 );
    p_comm_block->steps[i].p_mask_buffer[0] = 0xFF;
    p_comm_block->steps[i].p_mask_buffer[1] = 0xFF;
    memset( &p_comm_block->steps[i].p_mask_buffer[2], 0, 32 );
  }
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //copy data to user-space structure
    for( i = 0; i < 16; i++ )
      copy_to_user( &p_read_sec_buf->data[i << 5], &p_comm_block->steps[i].p_in_buffer[2], 32 );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl read_sec_buf on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );
  
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get CF sector count
//returns the number of sectors on compact flash
static int am_usb_file_ioctl_get_cf_sec_cnt( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_cf_sec_cnt * p_get_cf_sec_cnt )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_get_cf_sec_cnt get_cf_sec_cnt;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_cf_sec_cnt, sizeof( struct t_am_usb_get_cf_sec_cnt ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x05" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  memcpy( p_comm_block->steps[0].p_out_buffer, "\x40", 1 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x40\0\0\0\0", 5 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0", 5 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    get_cf_sec_cnt.sec_cnt = __be32_to_cpu( *(__u32 *)(&p_comm_block->steps[0].p_in_buffer[1]) );
    //copy structure to user-space
    copy_to_user( p_get_cf_sec_cnt, &get_cf_sec_cnt, sizeof( struct t_am_usb_get_cf_sec_cnt ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_cf_sec_cnt on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//write CF sector
//writes the sector buffer to a sector on compact flash
static int am_usb_file_ioctl_write_cf( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_write_cf * p_write_cf )
{
  struct t_am_usb_write_cf write_cf;
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_write_cf, sizeof( struct t_am_usb_write_cf ) ) ) //WRITE implies READ
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &write_cf, p_write_cf, sizeof( struct t_am_usb_write_cf ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x05", "\x06" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x42;
  *(__u32 *)(&p_comm_block->steps[0].p_out_buffer[1]) = __cpu_to_be32( write_cf.sec_no );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x42\0\0\0\0\0", 6 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0\0", 6 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    write_cf.sec_no = __be32_to_cpu( *(__u32 *)(&p_comm_block->steps[0].p_in_buffer[1]) );
    write_cf.err = p_comm_block->steps[0].p_in_buffer[5];
    //copy structure to user-space
    copy_to_user( p_write_cf, &write_cf, sizeof( struct t_am_usb_write_cf ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl write_cf on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//read CF sector
//reads a sector on compact flash into sector buffer
static int am_usb_file_ioctl_read_cf( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_read_cf * p_read_cf )
{
  struct t_am_usb_read_cf read_cf;
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_read_cf, sizeof( struct t_am_usb_read_cf ) ) ) //WRITE implies READ
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &read_cf, p_read_cf, sizeof( struct t_am_usb_read_cf ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x05", "\x06" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x43;
  *(__u32 *)(&p_comm_block->steps[0].p_out_buffer[1]) = __cpu_to_be32( read_cf.sec_no );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x43\0\0\0\0\0", 6 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0\0", 6 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    read_cf.sec_no = __be32_to_cpu( *(__u32 *)(&p_comm_block->steps[0].p_in_buffer[1]) );
    read_cf.err = p_comm_block->steps[0].p_in_buffer[5];
    //copy structure to user-space
    copy_to_user( p_read_cf, &read_cf, sizeof( struct t_am_usb_read_cf ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl read_cf on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get internal sector count
//returns the number of sectors in internal flash
static int am_usb_file_ioctl_get_if_sec_cnt( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_if_sec_cnt * p_get_if_sec_cnt )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_get_if_sec_cnt get_if_sec_cnt;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_if_sec_cnt, sizeof( struct t_am_usb_get_if_sec_cnt ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x05" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  memcpy( p_comm_block->steps[0].p_out_buffer, "\x44", 1 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x44\0\0\0\0", 5 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0", 5 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    get_if_sec_cnt.sec_cnt = __be32_to_cpu( *(__u32 *)(&p_comm_block->steps[0].p_in_buffer[1]) );
    //copy structure to user-space
    copy_to_user( p_get_if_sec_cnt, &get_if_sec_cnt, sizeof( struct t_am_usb_get_if_sec_cnt ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_if_sec_cnt on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//read internal sector
//reads a sector in internal flash into sector buffer
static int am_usb_file_ioctl_read_if( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_read_if * p_read_if )
{
  struct t_am_usb_read_if read_if;
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_read_if, sizeof( struct t_am_usb_read_if ) ) ) //WRITE implies READ
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &read_if, p_read_if, sizeof( struct t_am_usb_read_if ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x05", "\x06" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x47;
  *(__u32 *)(&p_comm_block->steps[0].p_out_buffer[1]) = __cpu_to_be32( read_if.sec_no );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x47\0\0\0\0\0", 6 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0\0", 6 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    read_if.sec_no = __be32_to_cpu( *(__u32 *)(&p_comm_block->steps[0].p_in_buffer[1]) );
    read_if.err = p_comm_block->steps[0].p_in_buffer[5];
    //copy structure to user-space
    copy_to_user( p_read_if, &read_if, sizeof( struct t_am_usb_read_if ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl read_if on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get compact flash flag
//returns if commands are executed from compact flash or from internal flash
static int am_usb_file_ioctl_get_cf_flag( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_cf_flag * p_get_cf_flag )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_get_cf_flag get_cf_flag;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_cf_flag, sizeof( struct t_am_usb_get_cf_flag ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  memcpy( p_comm_block->steps[0].p_out_buffer, "\x48", 1 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x48\0", 2 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0", 2 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    get_cf_flag.cf_flag = p_comm_block->steps[0].p_in_buffer[1] == 0 ? 0 : 1;
    //copy structure to user-space
    copy_to_user( p_get_cf_flag, &get_cf_flag, sizeof( struct t_am_usb_get_cf_flag ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_cf_flag on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//load variables
//loads the variables with 256 bytes of data
static int am_usb_file_ioctl_load_vars( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_load_vars * p_load_vars )
{
  struct t_am_usb_comm_block * p_comm_block;
  int i, ret_val;

  //test pointer
  if( verify_area( VERIFY_READ, p_load_vars, sizeof( struct t_am_usb_load_vars ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 8, "\x22\x22\x22\x22\x22\x22\x22\x22",
                                           "\x02\x02\x02\x02\x02\x02\x02\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  for( i = 0; i < 8; i++ )
  {
    p_comm_block->steps[i].p_out_buffer[0] = 0x50;
    p_comm_block->steps[i].p_out_buffer[1] = i;
    copy_from_user( &p_comm_block->steps[i].p_out_buffer[2], &p_load_vars->vars[i << 5], 32 );
    p_comm_block->steps[i].p_filter_buffer[0] = 0x50;
    p_comm_block->steps[i].p_filter_buffer[1] = i;
    p_comm_block->steps[i].p_mask_buffer[0] = 0xFF;
    p_comm_block->steps[i].p_mask_buffer[1] = 0xFF;
  }
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl load_vars on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//read variables
//reads 256 bytes of data from the variables
static int am_usb_file_ioctl_read_vars( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_read_vars * p_read_vars )
{
  struct t_am_usb_comm_block * p_comm_block;
  int i, ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_read_vars, sizeof( struct t_am_usb_read_vars ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 8, "\x02\x02\x02\x02\x02\x02\x02\x02",
                                           "\x22\x22\x22\x22\x22\x22\x22\x22" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  for( i = 0; i < 8; i++ )
  {
    p_comm_block->steps[i].p_out_buffer[0] = 0x51;
    p_comm_block->steps[i].p_out_buffer[1] = i;
    p_comm_block->steps[i].p_filter_buffer[0] = 0x51;
    p_comm_block->steps[i].p_filter_buffer[1] = i;
    memset( &p_comm_block->steps[i].p_filter_buffer[2], 0, 32 );
    p_comm_block->steps[i].p_mask_buffer[0] = 0xFF;
    p_comm_block->steps[i].p_mask_buffer[1] = 0xFF;
    memset( &p_comm_block->steps[i].p_mask_buffer[2], 0, 32 );
  }
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //copy data to user-space structure
    for( i = 0; i < 8; i++ )
      copy_to_user( &p_read_vars->vars[i << 5], &p_comm_block->steps[i].p_in_buffer[2], 32 );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl read_vars on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );
  
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//load stack
//loads the stack with 128 bytes of data
static int am_usb_file_ioctl_load_stack( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_load_stack * p_load_stack )
{
  struct t_am_usb_comm_block * p_comm_block;
  int i, ret_val;

  //test pointer
  if( verify_area( VERIFY_READ, p_load_stack, sizeof( struct t_am_usb_load_stack ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 4, "\x22\x22\x22\x22",
                                           "\x02\x02\x02\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  for( i = 0; i < 4; i++ )
  {
    p_comm_block->steps[i].p_out_buffer[0] = 0x54;
    p_comm_block->steps[i].p_out_buffer[1] = i;
    copy_from_user( &p_comm_block->steps[i].p_out_buffer[2], &p_load_stack->data[i << 5], 32 );
    p_comm_block->steps[i].p_filter_buffer[0] = 0x54;
    p_comm_block->steps[i].p_filter_buffer[1] = i;
    p_comm_block->steps[i].p_mask_buffer[0] = 0xFF;
    p_comm_block->steps[i].p_mask_buffer[1] = 0xFF;
  }
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl load_stack on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//read stack
//reads 128 bytes of data from the stack
static int am_usb_file_ioctl_read_stack( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_read_stack * p_read_stack )
{
  struct t_am_usb_comm_block * p_comm_block;
  int i, ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_read_stack, sizeof( struct t_am_usb_read_stack ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 4, "\x02\x02\x02\x02",
                                           "\x22\x22\x22\x22" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  for( i = 0; i < 4; i++ )
  {
    p_comm_block->steps[i].p_out_buffer[0] = 0x55;
    p_comm_block->steps[i].p_out_buffer[1] = i;
    p_comm_block->steps[i].p_filter_buffer[0] = 0x55;
    p_comm_block->steps[i].p_filter_buffer[1] = i;
    memset( &p_comm_block->steps[i].p_filter_buffer[2], 0, 32 );
    p_comm_block->steps[i].p_mask_buffer[0] = 0xFF;
    p_comm_block->steps[i].p_mask_buffer[1] = 0xFF;
    memset( &p_comm_block->steps[i].p_mask_buffer[2], 0, 32 );
  }
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //copy data to user-space structure
    for( i = 0; i < 4; i++ )
      copy_to_user( &p_read_stack->data[i << 5], &p_comm_block->steps[i].p_in_buffer[2], 32 );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl read_stack on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );
  
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//set stack pointer
static int am_usb_file_ioctl_set_stack_ptr( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_set_stack_ptr * p_set_stack_ptr )
{
  struct t_am_usb_set_stack_ptr set_stack_ptr;
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //test pointer
  if( verify_area( VERIFY_READ, p_set_stack_ptr, sizeof( struct t_am_usb_set_stack_ptr ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &set_stack_ptr, p_set_stack_ptr, sizeof( struct t_am_usb_set_stack_ptr ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x02", "\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x56;
  p_comm_block->steps[0].p_out_buffer[1] = set_stack_ptr.stack_ptr;
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x56\0", 2 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0", 2 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    set_stack_ptr.stack_ptr = p_comm_block->steps[0].p_in_buffer[1];
    //copy structure to user-space
    copy_to_user( p_set_stack_ptr, &set_stack_ptr, sizeof( struct t_am_usb_set_stack_ptr ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl set_stack_ptr on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );
  
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get stack pointer
static int am_usb_file_ioctl_get_stack_ptr( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_stack_ptr * p_get_stack_ptr )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_get_stack_ptr get_stack_ptr;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_stack_ptr, sizeof( struct t_am_usb_get_stack_ptr ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  memcpy( p_comm_block->steps[0].p_out_buffer, "\x57", 1 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x57\0", 2 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0", 2 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    get_stack_ptr.stack_ptr = p_comm_block->steps[0].p_in_buffer[1];
    //copy structure to user-space
    copy_to_user( p_get_stack_ptr, &get_stack_ptr, sizeof( struct t_am_usb_get_stack_ptr ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_stack_ptr on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );
  
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get and set key state
//gets key state and sets key state afterwards
static int am_usb_file_ioctl_get_set_keys( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_set_keys * p_get_set_keys )
{
  struct t_am_usb_get_set_keys get_set_keys;
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_set_keys, sizeof( struct t_am_usb_get_set_keys ) ) ) //WRITE implies READ
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &get_set_keys, p_get_set_keys, sizeof( struct t_am_usb_get_set_keys ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x05", "\x04" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x60;
  p_comm_block->steps[0].p_out_buffer[1] = get_set_keys.in_pressed;
  p_comm_block->steps[0].p_out_buffer[2] = get_set_keys.in_pressed_mask;
  p_comm_block->steps[0].p_out_buffer[3] = get_set_keys.in_released;
  p_comm_block->steps[0].p_out_buffer[4] = get_set_keys.in_released_mask;
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x60\0\0\0", 4 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0", 4 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    get_set_keys.state = p_comm_block->steps[0].p_in_buffer[1];
    get_set_keys.pressed = p_comm_block->steps[0].p_in_buffer[2];
    get_set_keys.released = p_comm_block->steps[0].p_in_buffer[3];
    //copy structure to user-space
    copy_to_user( p_get_set_keys, &get_set_keys, sizeof( struct t_am_usb_get_set_keys ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );
  
  //debug message
  am_usb_debugf( "ioctl get_set_keys on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );
  
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get and set SNES0 key state
//gets SNES0 key state and sets SNES0 key state afterwards
static int am_usb_file_ioctl_get_set_snes0( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_set_snes0 * p_get_set_snes0 )
{
  struct t_am_usb_get_set_snes0 get_set_snes0;
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_set_snes0, sizeof( struct t_am_usb_get_set_snes0 ) ) ) //WRITE implies READ
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &get_set_snes0, p_get_set_snes0, sizeof( struct t_am_usb_get_set_snes0 ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x09", "\x07" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x62;
  *(__u16 *)(&p_comm_block->steps[0].p_out_buffer[1]) = __cpu_to_be16( get_set_snes0.in_pressed );
  *(__u16 *)(&p_comm_block->steps[0].p_out_buffer[3]) = __cpu_to_be16( get_set_snes0.in_pressed_mask );
  *(__u16 *)(&p_comm_block->steps[0].p_out_buffer[5]) = __cpu_to_be16( get_set_snes0.in_released );
  *(__u16 *)(&p_comm_block->steps[0].p_out_buffer[7]) = __cpu_to_be16( get_set_snes0.in_released_mask );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x62\0\0\0\0\0\0", 7 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0\0\0", 7 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    get_set_snes0.state = __be16_to_cpu( *(__u16 *)(&p_comm_block->steps[0].p_in_buffer[1]) );
    get_set_snes0.pressed = __be16_to_cpu( *(__u16 *)(&p_comm_block->steps[0].p_in_buffer[3]) );
    get_set_snes0.released = __be16_to_cpu( *(__u16 *)(&p_comm_block->steps[0].p_in_buffer[5]) );
    //copy structure to user-space
    copy_to_user( p_get_set_snes0, &get_set_snes0, sizeof( struct t_am_usb_get_set_snes0 ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_set_snes0 on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );
  
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get and set SNES1 key state
//gets SNES1 key state and sets SNES1 key state afterwards
static int am_usb_file_ioctl_get_set_snes1( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_set_snes1 * p_get_set_snes1 )
{
  struct t_am_usb_get_set_snes1 get_set_snes1;
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_set_snes1, sizeof( struct t_am_usb_get_set_snes1 ) ) ) //WRITE implies READ
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &get_set_snes1, p_get_set_snes1, sizeof( struct t_am_usb_get_set_snes1 ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x09", "\x07" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x63;
  *(__u16 *)(&p_comm_block->steps[0].p_out_buffer[1]) = __cpu_to_be16( get_set_snes1.in_pressed );
  *(__u16 *)(&p_comm_block->steps[0].p_out_buffer[3]) = __cpu_to_be16( get_set_snes1.in_pressed_mask );
  *(__u16 *)(&p_comm_block->steps[0].p_out_buffer[5]) = __cpu_to_be16( get_set_snes1.in_released );
  *(__u16 *)(&p_comm_block->steps[0].p_out_buffer[7]) = __cpu_to_be16( get_set_snes1.in_released_mask );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x63\0\0\0\0\0\0", 7 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0\0\0", 7 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    get_set_snes1.state = __be16_to_cpu( *(__u16 *)(&p_comm_block->steps[0].p_in_buffer[1]) );
    get_set_snes1.pressed = __be16_to_cpu( *(__u16 *)(&p_comm_block->steps[0].p_in_buffer[3]) );
    get_set_snes1.released = __be16_to_cpu( *(__u16 *)(&p_comm_block->steps[0].p_in_buffer[5]) );
    //copy structure to user-space
    copy_to_user( p_get_set_snes1, &get_set_snes1, sizeof( struct t_am_usb_get_set_snes1 ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );
  
  //debug message
  am_usb_debugf( "ioctl get_set_snes1 on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );
  
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//set time
//sets the time of the ARCADEmini closk
static int am_usb_file_ioctl_set_time( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_set_time * p_set_time )
{
  struct t_am_usb_set_time set_time;
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;

  //test pointer
  if( verify_area( VERIFY_READ, p_set_time, sizeof( struct t_am_usb_set_time ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &set_time, p_set_time, sizeof( struct t_am_usb_set_time ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x04", "\x01" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x70;
  p_comm_block->steps[0].p_out_buffer[1] = set_time.sec;
  p_comm_block->steps[0].p_out_buffer[2] = set_time.min;
  p_comm_block->steps[0].p_out_buffer[3] = set_time.hour;
  p_comm_block->steps[0].p_filter_buffer[0] = 0x70;
  p_comm_block->steps[0].p_mask_buffer[0] = 0xFF;
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl set_time on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );
  
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get time
//gets the time from the ARCADEmini clock
static int am_usb_file_ioctl_get_time( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_time * p_get_time )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_get_time get_time;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_time, sizeof( struct t_am_usb_get_time ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x04" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  memcpy( p_comm_block->steps[0].p_out_buffer, "\x71", 1 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x71\0\0\0", 4 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0", 4 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    get_time.sec = p_comm_block->steps[0].p_in_buffer[1];
    get_time.min = p_comm_block->steps[0].p_in_buffer[2];
    get_time.hour = p_comm_block->steps[0].p_in_buffer[3];
    //copy structure to user-space
    copy_to_user( p_get_time, &get_time, sizeof( struct t_am_usb_get_time ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_time on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );
  
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get firmware version
//returns the firmware version of ARCADEmini
static int am_usb_file_ioctl_get_fw_version( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_fw_version * p_get_fw_version )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_get_fw_version get_fw_version;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_fw_version, sizeof( struct t_am_usb_get_fw_version ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x07" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  memcpy( p_comm_block->steps[0].p_out_buffer, "\x80", 1 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x80\0\0\0\0\0\0", 7 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0\0\0", 7 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    get_fw_version.major = p_comm_block->steps[0].p_in_buffer[1];
    get_fw_version.minor = p_comm_block->steps[0].p_in_buffer[2];
    get_fw_version.revision = p_comm_block->steps[0].p_in_buffer[3];
    get_fw_version.year = p_comm_block->steps[0].p_in_buffer[4];
    get_fw_version.month = p_comm_block->steps[0].p_in_buffer[5];
    get_fw_version.day = p_comm_block->steps[0].p_in_buffer[6];
    //copy structure to user-space
    copy_to_user( p_get_fw_version, &get_fw_version, sizeof( struct t_am_usb_get_fw_version ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_fw_version on file %08X (%d, %d): %d", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, ret_val );
  
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get ID
//gets the ID of this ARCADEmini
static int am_usb_file_ioctl_get_id( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_id * p_get_id )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_get_id get_id;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_id, sizeof( struct t_am_usb_get_id ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x12" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  memcpy( p_comm_block->steps[0].p_out_buffer, "\x84", 1 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x84\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 18 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 18 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    memcpy( get_id.id, &p_comm_block->steps[0].p_in_buffer[1], 16 );
    get_id.err = p_comm_block->steps[0].p_in_buffer[17];
    //copy structure to user-space
    copy_to_user( p_get_id, &get_id, sizeof( struct t_am_usb_get_id ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_id on file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//set ID
//sets the ID of this ARCADEmini
static int am_usb_file_ioctl_set_id( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_set_id * p_set_id )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_set_id set_id;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_set_id, sizeof( struct t_am_usb_set_id ) ) ) //WRITE implies READ
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &set_id, p_set_id, sizeof( struct t_am_usb_set_id ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x11", "\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x85;
  memcpy( &p_comm_block->steps[0].p_out_buffer[1], set_id.id, 16 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x85\0", 2 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0", 2 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    set_id.err = p_comm_block->steps[0].p_in_buffer[1];
    //copy structure to user-space
    copy_to_user( p_set_id, &set_id, sizeof( struct t_am_usb_set_id ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl set_id on file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get output styling
//gets the output styling
static int am_usb_file_ioctl_get_out_style( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_out_style * p_get_out_style )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_get_out_style get_out_style;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_out_style, sizeof( struct t_am_usb_get_out_style ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x03" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  memcpy( p_comm_block->steps[0].p_out_buffer, "\x88", 1 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x88\0\0", 3 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0", 3 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    get_out_style.style = p_comm_block->steps[0].p_in_buffer[1];
    get_out_style.err = p_comm_block->steps[0].p_in_buffer[2];
    //copy structure to user-space
    copy_to_user( p_get_out_style, &get_out_style, sizeof( struct t_am_usb_get_out_style ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_out_style on file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//set output style
//sets the output style
static int am_usb_file_ioctl_set_out_style( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_set_out_style * p_set_out_style )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_set_out_style set_out_style;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_set_out_style, sizeof( struct t_am_usb_set_out_style ) ) ) //WRITE implies READ
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &set_out_style, p_set_out_style, sizeof( struct t_am_usb_set_out_style ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x02", "\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x89;
  p_comm_block->steps[0].p_out_buffer[1] = set_out_style.style;
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x89\0", 2 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0", 2 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    set_out_style.err = p_comm_block->steps[0].p_in_buffer[1];
    //copy structure to user-space
    copy_to_user( p_set_out_style, &set_out_style, sizeof( struct t_am_usb_set_out_style ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl set_out_style on file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get output timing
//gets the output timing
static int am_usb_file_ioctl_get_out_timing( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_out_timing * p_get_out_timing )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_get_out_timing get_out_timing;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_out_timing, sizeof( struct t_am_usb_get_out_timing ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x11" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  memcpy( p_comm_block->steps[0].p_out_buffer, "\x8A", 1 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x8A\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 17 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 17 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    memcpy( get_out_timing.timing, &p_comm_block->steps[0].p_in_buffer[1], 15 );
    get_out_timing.err = p_comm_block->steps[0].p_in_buffer[16];
    //copy structure to user-space
    copy_to_user( p_get_out_timing, &get_out_timing, sizeof( struct t_am_usb_get_out_timing ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_out_timing on file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//set output timing
//sets the output timing
static int am_usb_file_ioctl_set_out_timing( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_set_out_timing * p_set_out_timing )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_set_out_timing set_out_timing;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_set_out_timing, sizeof( struct t_am_usb_set_out_timing ) ) ) //WRITE implies READ
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &set_out_timing, p_set_out_timing, sizeof( struct t_am_usb_set_out_timing ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x10", "\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x8B;
  memcpy( &p_comm_block->steps[0].p_out_buffer[1], set_out_timing.timing, 15 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x8B\0", 2 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0", 2 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    set_out_timing.err = p_comm_block->steps[0].p_in_buffer[1];
    //copy structure to user-space
    copy_to_user( p_set_out_timing, &set_out_timing, sizeof( struct t_am_usb_set_out_timing ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl set_out_timing on file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get output greyscale mapping
//gets the output greyscale mapping
static int am_usb_file_ioctl_get_out_greymap( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_out_greymap * p_get_out_greymap )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_get_out_greymap get_out_greymap;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_out_greymap, sizeof( struct t_am_usb_get_out_greymap ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x11" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  memcpy( p_comm_block->steps[0].p_out_buffer, "\x8C", 1 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x8C\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 17 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 17 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    memcpy( get_out_greymap.greymap, &p_comm_block->steps[0].p_in_buffer[1], 15 );
    get_out_greymap.err = p_comm_block->steps[0].p_in_buffer[16];
    //copy structure to user-space
    copy_to_user( p_get_out_greymap, &get_out_greymap, sizeof( struct t_am_usb_get_out_greymap ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl get_out_greymap on file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//set output greymap
//sets the output greymap
static int am_usb_file_ioctl_set_out_greymap( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_set_out_greymap * p_set_out_greymap )
{
  struct t_am_usb_comm_block * p_comm_block;
  int ret_val;
  struct t_am_usb_set_out_greymap set_out_greymap;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_set_out_greymap, sizeof( struct t_am_usb_set_out_greymap ) ) ) //WRITE implies READ
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &set_out_greymap, p_set_out_greymap, sizeof( struct t_am_usb_set_out_greymap ) );

  //allocate communication block
  p_comm_block = am_usb_comm_block_new( 1, "\x10", "\x02" );
  if( p_comm_block == NULL )
  {
    //unlock device for file operations
    up( &p_dev->file_lock );
    return -ENOMEM;
  }
  //fill communication block
  p_comm_block->steps[0].p_out_buffer[0] = 0x8D;
  memcpy( &p_comm_block->steps[0].p_out_buffer[1], set_out_greymap.greymap, 15 );
  memcpy( p_comm_block->steps[0].p_filter_buffer, "\x8D\0", 2 );
  memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0", 2 );
  //process commmunication block
  ret_val = am_usb_comm_block_do( p_dev, p_comm_block );

  //success
  if( ret_val == 0 )
  {
    //fill local structure
    set_out_greymap.err = p_comm_block->steps[0].p_in_buffer[1];
    //copy structure to user-space
    copy_to_user( p_set_out_greymap, &set_out_greymap, sizeof( struct t_am_usb_set_out_greymap ) );
  }

  //free communication block
  am_usb_comm_block_free( p_comm_block );

  //debug message
  am_usb_debugf( "ioctl set_out_greymap on file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return with error code
  return ret_val;
}

//get version information
//returns version information about the driver
static int am_usb_file_ioctl_get_version( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_version * p_get_version )
{
  struct t_am_usb_get_version get_version;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_version, sizeof( struct t_am_usb_get_version ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //fill local structure
  get_version.major = am_usb_ver_major;
  get_version.minor = am_usb_ver_minor;
  get_version.revision = am_usb_ver_revision;
  get_version.year = am_usb_ver_year;
  get_version.month = am_usb_ver_month;
  get_version.day = am_usb_ver_day;

  //copy structure to user-space
  copy_to_user( p_get_version, &get_version, sizeof( struct t_am_usb_get_version ) );

  //debug message
  am_usb_debugf( "ioctl get_version on file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return success
  return 0;
}

//set mode
//sets the driver mode for normal file operations
static int am_usb_file_ioctl_set_mode( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_set_mode * p_set_mode )
{
  struct t_am_usb_set_mode set_mode;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_set_mode, sizeof( struct t_am_usb_set_mode ) ) ) //WRITE implies READ
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //copy structure from user-space
  copy_from_user( &set_mode, p_set_mode, sizeof( struct t_am_usb_set_mode ) );

  //lock device
  down( &p_dev->lock );
  //get new mode from local structure
  switch( set_mode.mode )
  {
    //known modes
    case am_usb_mode_null:
    case am_usb_mode_stream:
    case am_usb_mode_cf:
      //accept new mode for this file
      p_file_priv->mode = set_mode.mode;
      break;
  }
  //get new default mode from local structure
  switch( set_mode.def_mode )
  {
    //known modes
    case am_usb_mode_null:
    case am_usb_mode_stream:
    case am_usb_mode_cf:
      //accept new default mode for this device
      p_dev->mode = set_mode.def_mode;
      break;
  }
  //fill local structure with new device mode and file mode
  set_mode.mode = p_file_priv->mode;
  set_mode.def_mode = p_dev->mode;
  //unlock device
  up( &p_dev->lock );

  //copy structure to user-space
  copy_to_user( p_set_mode, &set_mode, sizeof( struct t_am_usb_set_mode ) );

  //debug message
  am_usb_debugf( "ioctl set_mode on file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return success
  return 0;
}

//get mode
//gets the driver mode for normal file operations
static int am_usb_file_ioctl_get_mode( struct file * p_file, struct t_am_usb_file_priv * p_file_priv, struct t_am_usb_dev * p_dev, struct t_am_usb_get_mode * p_get_mode )
{
  struct t_am_usb_get_mode get_mode;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_get_mode, sizeof( struct t_am_usb_get_mode ) ) )
    //pointer not okay
    return -EFAULT;

  //lock device for file operations
  down( &p_dev->file_lock );

  //lock device
  down( &p_dev->lock );
  //fill local structure with current mode and default mode
  get_mode.mode = p_file_priv->mode;
  get_mode.def_mode = p_dev->mode;
  //unlock device
  up( &p_dev->lock );

  //copy structure to user-space
  copy_to_user( p_get_mode, &get_mode, sizeof( struct t_am_usb_get_mode ) );

  //debug message
  am_usb_debugf( "ioctl get_mode on file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return success
  return 0;
}

//an userspace process is doing an ioctl on one of the devices
static int am_usb_file_ioctl( struct inode * p_inode, struct file * p_file, unsigned int ioctl_no, unsigned long ioctl_param )
{
  struct t_am_usb_file_priv * p_file_priv;
  struct t_am_usb_dev * p_dev;

  //get private file
  p_file_priv = (struct t_am_usb_file_priv *)p_file->private_data;
  //get private device
  p_dev = p_file_priv->p_dev;
  //no pointer to device
  if( p_dev == NULL )
    //so there is no device
    return -ENODEV;

  //choose ioctl
  switch( ioctl_no )
  {
    //halt ARCADEmini
    case am_usb_ioctl_no_halt:
      return am_usb_file_ioctl_halt( p_file, p_file_priv, p_dev );
    //run ARCADEmini
    case am_usb_ioctl_no_run:
      return am_usb_file_ioctl_run( p_file, p_file_priv, p_dev );
    //step ARCADEmini
    case am_usb_ioctl_no_step:
      return am_usb_file_ioctl_step( p_file, p_file_priv, p_dev );
    //set command pointer
    case am_usb_ioctl_no_set_cmd_ptr:
      return am_usb_file_ioctl_set_cmd_ptr( p_file, p_file_priv, p_dev, (struct t_am_usb_set_cmd_ptr *)ioctl_param );
    //get command pointer
    case am_usb_ioctl_no_get_cmd_ptr:
      return am_usb_file_ioctl_get_cmd_ptr( p_file, p_file_priv, p_dev, (struct t_am_usb_get_cmd_ptr *)ioctl_param );
    //load pixel buffer
    case am_usb_ioctl_no_load_pix_buf:
      return am_usb_file_ioctl_load_pix_buf( p_file, p_file_priv, p_dev, (struct t_am_usb_load_pix_buf *)ioctl_param );
    //read pixel buffer
    case am_usb_ioctl_no_read_pix_buf:
      return am_usb_file_ioctl_read_pix_buf( p_file, p_file_priv, p_dev, (struct t_am_usb_read_pix_buf *)ioctl_param );
    //load overlay pixel buffer
    case am_usb_ioctl_no_load_ovp_buf:
      return am_usb_file_ioctl_load_ovp_buf( p_file, p_file_priv, p_dev, (struct t_am_usb_load_ovp_buf *)ioctl_param );
    //read overlay pixel buffer
    case am_usb_ioctl_no_read_ovp_buf:
      return am_usb_file_ioctl_read_ovp_buf( p_file, p_file_priv, p_dev, (struct t_am_usb_read_ovp_buf *)ioctl_param );
    //swap display buffers
    case am_usb_ioctl_no_swap_bufs:
      return am_usb_file_ioctl_swap_bufs( p_file, p_file_priv, p_dev );
    //swap display buffers and copy image
    case am_usb_ioctl_no_swap_bufs_cp:
      return am_usb_file_ioctl_swap_bufs_cp( p_file, p_file_priv, p_dev );
    //load sector buffer
    case am_usb_ioctl_no_load_sec_buf:
      return am_usb_file_ioctl_load_sec_buf( p_file, p_file_priv, p_dev, (struct t_am_usb_load_sec_buf *)ioctl_param );
    //read sector buffer
    case am_usb_ioctl_no_read_sec_buf:
      return am_usb_file_ioctl_read_sec_buf( p_file, p_file_priv, p_dev, (struct t_am_usb_read_sec_buf *)ioctl_param );
    //get CF sector count
    case am_usb_ioctl_no_get_cf_sec_cnt:
      return am_usb_file_ioctl_get_cf_sec_cnt( p_file, p_file_priv, p_dev, (struct t_am_usb_get_cf_sec_cnt *)ioctl_param );
    //write CF sector
    case am_usb_ioctl_no_write_cf:
      return am_usb_file_ioctl_write_cf( p_file, p_file_priv, p_dev, (struct t_am_usb_write_cf *)ioctl_param );
    //read CF sector
    case am_usb_ioctl_no_read_cf:
      return am_usb_file_ioctl_read_cf( p_file, p_file_priv, p_dev, (struct t_am_usb_read_cf *)ioctl_param );
    //get internal sector count
    case am_usb_ioctl_no_get_if_sec_cnt:
      return am_usb_file_ioctl_get_if_sec_cnt( p_file, p_file_priv, p_dev, (struct t_am_usb_get_if_sec_cnt *)ioctl_param );
    //read internal sector
    case am_usb_ioctl_no_read_if:
      return am_usb_file_ioctl_read_if( p_file, p_file_priv, p_dev, (struct t_am_usb_read_if *)ioctl_param );
    //get compact flash flag
    case am_usb_ioctl_no_get_cf_flag:
      return am_usb_file_ioctl_get_cf_flag( p_file, p_file_priv, p_dev, (struct t_am_usb_get_cf_flag *)ioctl_param );
    //load variables
    case am_usb_ioctl_no_load_vars:
      return am_usb_file_ioctl_load_vars( p_file, p_file_priv, p_dev, (struct t_am_usb_load_vars *)ioctl_param );
    //read variables
    case am_usb_ioctl_no_read_vars:
      return am_usb_file_ioctl_read_vars( p_file, p_file_priv, p_dev, (struct t_am_usb_read_vars *)ioctl_param );
    //load stack
    case am_usb_ioctl_no_load_stack:
      return am_usb_file_ioctl_load_stack( p_file, p_file_priv, p_dev, (struct t_am_usb_load_stack *)ioctl_param );
    //read stack
    case am_usb_ioctl_no_read_stack:
      return am_usb_file_ioctl_read_stack( p_file, p_file_priv, p_dev, (struct t_am_usb_read_stack *)ioctl_param );
    //set stack pointer
    case am_usb_ioctl_no_set_stack_ptr:
      return am_usb_file_ioctl_set_stack_ptr( p_file, p_file_priv, p_dev, (struct t_am_usb_set_stack_ptr *)ioctl_param );
    //get stack pointer
    case am_usb_ioctl_no_get_stack_ptr:
      return am_usb_file_ioctl_get_stack_ptr( p_file, p_file_priv, p_dev, (struct t_am_usb_get_stack_ptr *)ioctl_param );
    //get and set key state
    case am_usb_ioctl_no_get_set_keys:
      return am_usb_file_ioctl_get_set_keys( p_file, p_file_priv, p_dev, (struct t_am_usb_get_set_keys *)ioctl_param );
    //get and set SNES0 key state
    case am_usb_ioctl_no_get_set_snes0:
      return am_usb_file_ioctl_get_set_snes0( p_file, p_file_priv, p_dev, (struct t_am_usb_get_set_snes0 *)ioctl_param );
    //get and set SNES1 key state
    case am_usb_ioctl_no_get_set_snes1:
      return am_usb_file_ioctl_get_set_snes1( p_file, p_file_priv, p_dev, (struct t_am_usb_get_set_snes1 *)ioctl_param );
    //set time
    case am_usb_ioctl_no_set_time:
      return am_usb_file_ioctl_set_time( p_file, p_file_priv, p_dev, (struct t_am_usb_set_time *)ioctl_param );
    //get time
    case am_usb_ioctl_no_get_time:
      return am_usb_file_ioctl_get_time( p_file, p_file_priv, p_dev, (struct t_am_usb_get_time *)ioctl_param );
    //get firmware version
    case am_usb_ioctl_no_get_fw_version:
      return am_usb_file_ioctl_get_fw_version( p_file, p_file_priv, p_dev, (struct t_am_usb_get_fw_version *)ioctl_param );
    //get ID
    case am_usb_ioctl_no_get_id:
      return am_usb_file_ioctl_get_id( p_file, p_file_priv, p_dev, (struct t_am_usb_get_id *)ioctl_param );
    //set ID
    case am_usb_ioctl_no_set_id:
      return am_usb_file_ioctl_set_id( p_file, p_file_priv, p_dev, (struct t_am_usb_set_id *)ioctl_param );
    //get output style
    case am_usb_ioctl_no_get_out_style:
      return am_usb_file_ioctl_get_out_style( p_file, p_file_priv, p_dev, (struct t_am_usb_get_out_style *)ioctl_param );
    //set output style
    case am_usb_ioctl_no_set_out_style:
      return am_usb_file_ioctl_set_out_style( p_file, p_file_priv, p_dev, (struct t_am_usb_set_out_style *)ioctl_param );
    //get output timing
    case am_usb_ioctl_no_get_out_timing:
      return am_usb_file_ioctl_get_out_timing( p_file, p_file_priv, p_dev, (struct t_am_usb_get_out_timing *)ioctl_param );
    //set output timing
    case am_usb_ioctl_no_set_out_timing:
      return am_usb_file_ioctl_set_out_timing( p_file, p_file_priv, p_dev, (struct t_am_usb_set_out_timing *)ioctl_param );
    //get output greyscale mapping
    case am_usb_ioctl_no_get_out_greymap:
      return am_usb_file_ioctl_get_out_greymap( p_file, p_file_priv, p_dev, (struct t_am_usb_get_out_greymap *)ioctl_param );
    //set output greyscale mapping
    case am_usb_ioctl_no_set_out_greymap:
      return am_usb_file_ioctl_set_out_greymap( p_file, p_file_priv, p_dev, (struct t_am_usb_set_out_greymap *)ioctl_param );
    //get version information
    case am_usb_ioctl_no_get_version:
      return am_usb_file_ioctl_get_version( p_file, p_file_priv, p_dev, (struct t_am_usb_get_version *)ioctl_param );
    //set mode
    case am_usb_ioctl_no_set_mode:
      return am_usb_file_ioctl_set_mode( p_file, p_file_priv, p_dev, (struct t_am_usb_set_mode *)ioctl_param );
    //get mode
    case am_usb_ioctl_no_get_mode:
      return am_usb_file_ioctl_get_mode( p_file, p_file_priv, p_dev, (struct t_am_usb_get_mode *)ioctl_param );
  }

  //unknown ioctl
  am_usb_debugf( "unknown ioctl %08X on file %08X (%d, %d)", ioctl_no, (int)p_file, USB_MAJOR, p_dev->dev_no_minor );
  return -EINVAL;
}

//an userspace process is opening one of the devices
static int am_usb_file_open( struct inode * p_inode, struct file * p_file )
{
  struct t_am_usb_file_priv * p_file_priv;
  struct t_am_usb_dev * p_dev;
  int sub_minor, first_user;
  struct t_am_usb_comm_block * p_comm_block;
  
  //get subminor device number
  sub_minor = MINOR( p_inode->i_rdev ) - am_usb_minor_base;
  if( sub_minor < 0  || sub_minor >= am_usb_minor_range )
    //device not one of our devices
    return -ENODEV;

  //lock device table
  down( &am_usb_dev_table_lock );
  //get device structure for this minor
  p_dev = am_usb_dev_table[sub_minor];
  //device not attached
  if( p_dev == NULL )
  {
    //unlock device table
    up( &am_usb_dev_table_lock );
    //device not available
    return -ENODEV;
  }

  //lock device for file operations
  down( &p_dev->file_lock );
  //lock device
  down( &p_dev->lock );
  
  //unlock device table
  up( &am_usb_dev_table_lock );

  //device was unplugged
  if( p_dev->p_usb_dev == NULL )
  {
    //unlock device
    up( &p_dev->lock );
    //unlock device for file operations
    up( &p_dev->file_lock );
    //no such device there
    return -ENODEV;
  }
  
  //create private file data
  p_file_priv = (struct t_am_usb_file_priv *)kmalloc( sizeof( struct t_am_usb_file_priv ), GFP_KERNEL );
  if( p_file_priv == NULL )
  {
    //unlock device
    up( &p_dev->lock );
    //unlock device for file operations
    up( &p_dev->file_lock );
    //no memory available
    return -ENOMEM;
  }

  //initialize private file data
  //(p_dev and p_next_file_priv are initialized a few lines down)
  //initialize waitqueue for poll-call
  init_waitqueue_head( &p_file_priv->poll_wait_queue );
  //use default mode of device as file mode
  p_file_priv->mode = p_dev->mode;
  //set current sector to first one
  p_file_priv->sec_no = 0;
 
  //unlock device table
  up( &am_usb_dev_table_lock );

  //increment our usage count
  first_user = p_dev->open_cnt == 0; //remember if this is our first user
  p_dev->open_cnt++;

  //save pointer to private file data in file
  p_file->private_data = (void *)p_file_priv;
  //save pointer to device in the file's private structure
  p_file_priv->p_dev = p_dev;
  //put private file data into list in device structure
  //the list stays consistent all the time
  //(reading in interrupt is allowed when using only p_next or only p_prev)
  p_file_priv->p_prev = p_dev->file_priv_list.p_prev;
  p_file_priv->p_next = &p_dev->file_priv_list;
  p_file_priv->p_next->p_prev = p_file_priv;
  p_file_priv->p_prev->p_next = p_file_priv;

  //debug message
  am_usb_debugf( "opened file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device
  up( &p_dev->lock );

  //stop command execution if first user is opening
  if( first_user )
  {
    p_comm_block = am_usb_comm_block_new( 1, "\x01", "\x01" );
    if( p_comm_block != NULL )
    {
      p_comm_block->steps[0].p_out_buffer[0] = 0x00;
      p_comm_block->steps[0].p_filter_buffer[0] = 0x00;
      p_comm_block->steps[0].p_mask_buffer[0] = 0xFF;
      am_usb_comm_block_do( p_dev, p_comm_block );
      am_usb_comm_block_free( p_comm_block );
    }
  }

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return success
  return 0;
}

//an userspace process is closing one of the devices
static int am_usb_file_release( struct inode * p_inode, struct file * p_file )
{
  struct t_am_usb_file_priv * p_file_priv;
  struct t_am_usb_dev * p_dev;
  struct t_am_usb_comm_block * p_comm_block;
  int last_user;

  //get private file
  p_file_priv = (struct t_am_usb_file_priv *)p_file->private_data;
  //get private device
  p_dev = p_file_priv->p_dev;
  //no pointer to device
  if( p_dev == NULL )
    //so there is no device
    return -ENODEV;

  //lock device table
  down( &am_usb_dev_table_lock );

  //lock device for file operations
  down( &p_dev->file_lock );
  //lock device
  down( &p_dev->lock );

  //wake up poll-waitqueue of file (in case that another thread is doing a poll)
  wake_up_interruptible( &p_file_priv->poll_wait_queue );

  //remove private file data from list in device structure
  //the list stays consistent all the time
  //(reading in interrupt is allowed when using only p_next or only p_prev)
  p_file_priv->p_next->p_prev = p_file_priv->p_prev;
  p_file_priv->p_prev->p_next = p_file_priv->p_next;

  //free private file data
  kfree( p_file_priv );

  //check if device is open
  if( p_dev->open_cnt <= 0 )
  {
    //unlock device
    up( &p_dev->lock );
    //unlock device for file operations
    up( &p_dev->file_lock );
    //unlock device table
    up( &am_usb_dev_table_lock );
    //there is no such device
    return -ENODEV;
  }

  //debug message
  am_usb_debugf( "closed file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //decrement usage count
  p_dev->open_cnt--;
  last_user = p_dev->open_cnt == 0; //remember if this is our last user

  //device was unplugged
  if( p_dev->p_usb_dev == NULL )
  {
    //device is opened by someone else
    if( p_dev->open_cnt > 0 )
    {
      //unlock the device
      up( &p_dev->lock );
      //unlock device for file operations
      up( &p_dev->file_lock );
    }
    //device is not opened by anyone else
    else
    {
      //unlock the device
      up( &p_dev->lock );
      //unlock device for file operations
      up( &p_dev->file_lock );
      //free the device (also frees buffers and URBs)
      //(and remove it from the device table)
      am_usb_dev_free( p_dev );
    }
    //unlock device table
    up( &am_usb_dev_table_lock );
    //success
    return 0;
  }

  //unlock device
  up( &p_dev->lock );

  //unlock device table
  up( &am_usb_dev_table_lock );

  //restart command execution if last user is closing
  if( last_user )
  {
    p_comm_block = am_usb_comm_block_new( 2, "\x07\x01", "\x07\x01" );
    if( p_comm_block != NULL )
    {
      memcpy( p_comm_block->steps[0].p_out_buffer, "\x08\0\0\0\0\x01\xFF", 7 ); //set command pointer to sector 0 offset 0x1FF
      memcpy( p_comm_block->steps[0].p_filter_buffer, "\x08\0\0\0\0\0\0", 7 );
      memcpy( p_comm_block->steps[0].p_mask_buffer, "\xFF\0\0\0\0\0\0", 7 );
      p_comm_block->steps[1].p_out_buffer[0] = 0x01; //start command execution
      p_comm_block->steps[1].p_filter_buffer[0] = 0x01;
      p_comm_block->steps[1].p_mask_buffer[0] = 0xFF;
      am_usb_comm_block_do( p_dev, p_comm_block );
      am_usb_comm_block_free( p_comm_block );
    }
  }

  //unlock device for file operations
  up( &p_dev->file_lock );

  //return success
  return 0;
}

//structure for supported file operations
static struct file_operations am_usb_file_operations =
{
  owner:   THIS_MODULE, //because of this, we don't need the MOD_INC/DEC_USE_COUNT macros
  llseek:  am_usb_file_llseek,
  read:    am_usb_file_read,
  write:   am_usb_file_write,
  readdir: NULL,
  poll:    am_usb_file_poll,
  ioctl:   am_usb_file_ioctl,
  mmap:    NULL,
  open:    am_usb_file_open,
  flush:   NULL,
  release: am_usb_file_release,
};



//*** USB device operations

//device is attached to USB
static void * am_usb_probe( struct usb_device * p_usb_dev, unsigned int interface_no, const struct usb_device_id * p_usb_dev_id )
{
  unsigned char sub_minor;
  struct t_am_usb_dev * p_dev;
  struct usb_interface * p_interface;
  struct usb_interface_descriptor * p_interface_descriptor;
  int endpoints_ok;

  //check if vendor id and product id match
  if( p_usb_dev->descriptor.idVendor != am_usb_id_vendor || p_usb_dev->descriptor.idProduct != am_usb_id_product )
    //device can't be handled
    //(this is not an error, it's just not our device)
    return NULL;

  //now, we are sure, that the device is a am_usb-device

  //lock device table
  down( &am_usb_dev_table_lock );
  //select a free subminor number
  for( sub_minor = 0; sub_minor < am_usb_minor_range; sub_minor++ )
    if( am_usb_dev_table[sub_minor] == NULL )
      break;
  //no free subminor found
  if( sub_minor >= am_usb_minor_range )
  {
    //print error message
    am_usb_error( "too much devices" );
    //unlock device table
    up( &am_usb_dev_table_lock );
    //device can't be handled
    return NULL;
  }

  //allocate buffer for private data
  p_dev = (struct t_am_usb_dev *)kmalloc( sizeof( struct t_am_usb_dev ), GFP_KERNEL );
  //error
  if( p_dev == NULL )
  {
    //print error message
    am_usb_error( "could not allocate buffer for private data" );
    //unlock device table
    up( &am_usb_dev_table_lock );
    //abort function
    return NULL;
  }

  //put device into device table
  am_usb_dev_table[sub_minor] = p_dev;

  //initialize locks
  init_MUTEX( &p_dev->file_lock );
  init_MUTEX( &p_dev->lock );
  //save parameters of this function
  p_dev->p_usb_dev = p_usb_dev;
  p_dev->interface_no = interface_no;
  p_dev->p_usb_dev_id = p_usb_dev_id;
  //save minor device number
  p_dev->dev_no_sub_minor = sub_minor;
  p_dev->dev_no_minor = am_usb_minor_base + p_dev->dev_no_sub_minor;
  //no buffers yet
  p_dev->p_out_buffer = NULL;
  p_dev->p_in_buffer = NULL;
  //no URBs yet
  p_dev->p_out_urb = NULL;
  p_dev->p_in_urb = NULL;
  //URBs not submitted/linked yet
  p_dev->out_urb_linked = 0;
  p_dev->in_urb_linked = 0;
  //device is not opened yet
  p_dev->open_cnt = 0;
  //no files have opened this device
  p_dev->file_priv_list.p_dev = NULL;
  p_dev->file_priv_list.p_next = &p_dev->file_priv_list;
  p_dev->file_priv_list.p_prev = &p_dev->file_priv_list;
  //no communication blocks yet
  p_dev->comm_block_list.p_next = &p_dev->comm_block_list;
  p_dev->comm_block_list.p_prev = &p_dev->comm_block_list;
  p_dev->comm_block_list.step_cnt = 0;
  p_dev->comm_block_list.steps = NULL;
  p_dev->comm_block_list.out_step = 0;
  p_dev->comm_block_list.in_step = 0;
  p_dev->comm_block_list.p_wait_queue = NULL;
  p_dev->comm_block_list.free_flag = 0;
  //timeout timer not active
  p_dev->timeout_timer_active = 0;
  //default mode is null device
  p_dev->mode = overall_default_mode;

  //get our interface of current configuration
  p_interface = &p_usb_dev->actconfig->interface[interface_no];
  //get interface descriptor
  p_interface_descriptor = &p_interface->altsetting[0];

  //check endpoints: we want: bulk out, bulk in
  endpoints_ok = 0;
  if( p_interface_descriptor->bNumEndpoints == 2 ) //exactly 2 endpoints
  {
    endpoints_ok = 1;
    //bulk out
    if( (p_interface_descriptor->endpoint[0].bmAttributes & 0x03) == 0x02 //bulk
      && (p_interface_descriptor->endpoint[0].bEndpointAddress & 0x80) == 0x00 ) //out
    {
      //get out parameters
      p_dev->out_endpoint_addr = p_interface_descriptor->endpoint[0].bEndpointAddress & 0x7F;
      p_dev->out_max_packet_size = p_interface_descriptor->endpoint[0].wMaxPacketSize;
    }
    else
    {
      //endpoints are not okay
      endpoints_ok = 0;
    }
    //bulk in
    if( (p_interface_descriptor->endpoint[1].bmAttributes & 0x03) == 0x02 //bulk
      && (p_interface_descriptor->endpoint[1].bEndpointAddress & 0x80) != 0x00 ) //in
    {
      //get in parameters
      p_dev->in_endpoint_addr = p_interface_descriptor->endpoint[1].bEndpointAddress & 0x7F;
      p_dev->in_max_packet_size = p_interface_descriptor->endpoint[1].wMaxPacketSize;
    }
    else
    {
      //endpoints are not okay
      endpoints_ok = 0;
    }
  }

  //endpoints are not okay
  if( !endpoints_ok )
  {
    //print error message
    am_usb_error( "device has got unexpected endpoints" );
    //free device (also frees buffers and URBs)
    am_usb_dev_free( p_dev );
    //unlock device table
    up( &am_usb_dev_table_lock );
    //abort function
    return NULL;
  }

  //allocate buffers
  p_dev->p_out_buffer = kmalloc( p_dev->out_max_packet_size, GFP_KERNEL );
  p_dev->p_in_buffer = kmalloc( p_dev->in_max_packet_size, GFP_KERNEL );
  if( p_dev->p_out_buffer == NULL || p_dev->p_in_buffer == NULL )
  {
    //print error message
    am_usb_error( "could not allocate buffers" );
    //free device (also frees buffers and URBs)
    am_usb_dev_free( p_dev );
    //unlock device table
    up( &am_usb_dev_table_lock );
    //abort function
    return NULL;
  }

  //allocate URB for bulk output
  p_dev->p_out_urb = usb_alloc_urb( 0 );
  if( p_dev->p_out_urb == NULL )
  {
    //print error message
    am_usb_error( "could not allocate bulk out URB" );
    //free device (also frees buffers and URBs)
    am_usb_dev_free( p_dev );
    //unlock device table
    up( &am_usb_dev_table_lock );
    //abort function
    return NULL;
  }

  //allocate URB for bulk input
  p_dev->p_in_urb = usb_alloc_urb( 0 );
  if( p_dev->p_in_urb == NULL )
  {
    //print error message
    am_usb_error( "could not allocate bulk in URB" );
    //free device (also frees buffers and URBs)
    am_usb_dev_free( p_dev );
    //unlock device table
    up( &am_usb_dev_table_lock );
    //abort function
    return NULL;
  }
  //set up data in bulk in URB
  FILL_BULK_URB( p_dev->p_in_urb, p_usb_dev, usb_rcvbulkpipe( p_usb_dev, p_dev->in_endpoint_addr ),
      p_dev->p_in_buffer, p_dev->in_max_packet_size, am_usb_in_callback,
      p_dev );
  //bulk in URB will be linked/submitted soon
  //(do this here because callback could perhaps be called from usb_submit_urb)
  p_dev->in_urb_linked = 1;
  //submit bulk in URB
  if( usb_submit_urb( p_dev->p_in_urb ) )
  {
    //bulk in URB is not linked/submitted
    p_dev->in_urb_linked = 0;
    //print error message
    am_usb_error( "could not submit bulk in URB" );
    //free device (also frees buffers and URBs)
    am_usb_dev_free( p_dev );
    //unlock device table
    up( &am_usb_dev_table_lock );
    //abort function
    return NULL;
  }
  //print info that bulk in was started
  am_usb_debugf( "bulk input started (max_len=%d)", p_dev->p_in_urb->transfer_buffer_length );

  //register devfs node
  snprintf( p_dev->devfs_node_name, sizeof( p_dev->devfs_node_name ), am_usb_devfs_node_name"%d", p_dev->dev_no_sub_minor );
  p_dev->h_devfs_node = devfs_register( usb_devfs_handle, p_dev->devfs_node_name, DEVFS_FL_DEFAULT, USB_MAJOR, p_dev->dev_no_minor,
    S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH, &am_usb_file_operations, NULL );

  //print info message
  am_usb_infof( "device attached (major %d, minor %d, devfs name %s)", USB_MAJOR, p_dev->dev_no_minor, p_dev->devfs_node_name );

  //unlock device table
  up( &am_usb_dev_table_lock );

  //return pointer to private data
  return (void *)p_dev;
}

//device is removed from USB
static void am_usb_disconnect( struct usb_device * p_usb_dev, void * p_device )
{
  struct t_am_usb_dev * p_dev;

  //get pointer to device
  p_dev = (struct t_am_usb_dev *)p_device;

  //lock device for file operations
  down( &p_dev->file_lock );
  //lock device table and device
  down( &am_usb_dev_table_lock );
  down( &p_dev->lock );

  //print info message
  am_usb_infof( "device disconnected (major %d, minor %d, devfs name %s)", USB_MAJOR, p_dev->dev_no_minor, p_dev->devfs_node_name );

  //unregister devfs node
  p_dev->devfs_node_name[0] = 0;
  devfs_unregister( p_dev->h_devfs_node );

  //device is opened by someone
  if( p_dev->open_cnt > 0 )
  {
    //remove pointer to usb device
    //(device is gone)
    p_dev->p_usb_dev = NULL;

    //wake up poll-waitqueues of open files (in case that a thread is doing a poll)
    struct t_am_usb_file_priv * p_file_priv;
    for( p_file_priv = p_dev->file_priv_list.p_next;
         p_file_priv != &p_dev->file_priv_list;
         p_file_priv = p_file_priv->p_next )
      wake_up_interruptible( &p_file_priv->poll_wait_queue );

    //unlock the device
    up( &p_dev->lock );
    //unlock device for file operations
    up( &p_dev->file_lock );
  }
  //device is not opened by anyone
  else
  {
    //unlock the device
    up( &p_dev->lock );
    //free the device (and remove it from the device table)
    am_usb_dev_free( p_device );
  }

  //unlock device table
  up( &am_usb_dev_table_lock );
}

//table of supported devices
static struct usb_device_id am_usb_id_table [] =
{
  { USB_DEVICE( am_usb_id_vendor, am_usb_id_product ) },
  { } //end of table
};

//structure for USB data and functions
static struct usb_driver am_usb_usb_driver =
{
  name:        "am_usb",
  probe:       am_usb_probe,
  disconnect:  am_usb_disconnect,
  fops:        &am_usb_file_operations,
  minor:       am_usb_minor_base,
  id_table:    am_usb_id_table,
};



//*** module init and exit

//module is being loaded
static int am_usb_init( )
{
  //local variables
  int i;
  
#ifdef DEBUG_MEM
  //initialize memory debug lock
  init_MUTEX( &my_alloc_lock );
#endif
  
  //short message
  am_usb_info( am_usb_name" ("am_usb_ver")" );

  //parse overall default mode
  if( strcmp( mode, "null" ) == 0 )
    overall_default_mode = am_usb_mode_null;
  else if( strcmp( mode, "stream" ) == 0 )
    overall_default_mode = am_usb_mode_stream;
  else if( strcmp( mode, "cf" ) == 0 )
    overall_default_mode = am_usb_mode_cf;
  else if( mode[0] != 0 )
  {
    //error, someone does not seem to know what he's doing :-)
    am_usb_errorf( "Aieee, unknown mode: %s !!!", mode );
    return -1;
  }

  //initialize table with pointers to private device data of devices
  for( i = 0; i < am_usb_minor_range; i++ )
    am_usb_dev_table[i] = NULL;
  //initialize the lock for this table
  init_MUTEX( &am_usb_dev_table_lock );

  //reset boolean flags for tasks to do in next am_usb_task
  //(am_usb_task is task for things that cannot be done in interrupt)
  //(use test_and_x to be smp safe)
  for( i = 0; i < am_usb_minor_range; i++ )
  {
    test_and_clear_bit( am_usb_task_flag_comm_block_send, &am_usb_task_flags[i] );
    test_and_clear_bit( am_usb_task_flag_comm_block_recv, &am_usb_task_flags[i] );
  }

  //register this driver with the USB subsystem
  i = usb_register( &am_usb_usb_driver );
  //an error occured
  if( i < 0 )
  {
    //error-message
    am_usb_errorf( "usb_register failed with %d", i );
    //init failed
    return -1;
  }

  //init succeeded
  am_usb_info( "loaded..." );
  return 0;
}

//module is being unloaded
static void am_usb_exit( )
{
  //deregister driver with the USB subsystem
  usb_deregister( &am_usb_usb_driver );

  //wait until task is executed (task may be scheduled)
  flush_scheduled_tasks( );

  //exit done
  am_usb_info( "unloaded..." );
}


//module init and exit procedure
module_init( am_usb_init );
module_exit( am_usb_exit );

//table with supported devices
MODULE_DEVICE_TABLE( usb, am_usb_id_table );

